# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["TestCscCommander"]

import numpy as np

from . import csc_commander

# All arrays for setArrays have length 5
ARR_LEN = 5


class TestCscCommander(csc_commander.CscCommander):
    """Control a Test CSC from the command line.

    Parameters
    ----------
    name : `str`
        SAL component name of CSC.
    index : `int`, optional
        SAL index of CSC.
    enable : `bool`, optional
        Enable the CSC (when the commander starts up)?
        Note: `amain` always supplies this argument.
    """

    def __init__(
        self,
        index,
        enable=False,
        exclude=None,
        fields_to_ignore=("ignored", "value", "priority"),
    ):
        super().__init__(
            name="Test",
            index=index,
            enable=enable,
            exclude_commands=("abort", "enterControl", "setValue"),
        )

        def asbool(val):
            """Cast an int or string representation of an int to a boolean."""
            return bool(int(val))

        self.array_field_types = dict(
            boolean0=asbool,
            byte0=np.uint8,
            short0=np.int16,
            int0=np.int32,
            long0=np.int32,
            longLong0=np.int64,
            octet0=np.uint8,
            unsignedShort0=np.uint16,
            unsignedInt0=np.uint32,
            unsignedLong0=np.uint32,
            float0=np.single,
            double0=np.double,
        )

        self.help_dict[
            "setArrays"
        ] = f"""field1=val1,val2,... field2=val1,va2,... (up to {ARR_LEN} values each)
          field: {" ".join(self.array_field_types)}
          for example: boolean0=1,2 int0=3,-4,5,6,7 float0=-42.3"""

    async def do_setArrays(self, args):
        kwargs = dict()
        for field_val in args:
            field, valstr = field_val.split("=")
            if field not in self.array_field_types:
                raise RuntimeError(f"Unknown field {field}")
            field_type = self.array_field_types[field]
            vals = [field_type(val) for val in valstr.split(",")]
            if len(vals) > ARR_LEN:
                raise RuntimeError(
                    f"Field {field} has more than {ARR_LEN} values: {vals}"
                )
            elif len(vals) < ARR_LEN:
                n_to_append = ARR_LEN - len(vals)
                vals += [field_type(0)] * n_to_append
            kwargs[field] = vals
        await self.remote.cmd_setArrays.set_start(**kwargs)



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["TestCsc"]

import asyncio
import string

import numpy as np

from .base_csc import State
from .configurable_csc import ConfigurableCsc
from . import __version__
from .config_schema import CONFIG_SCHEMA


class TestCsc(ConfigurableCsc):
    """A simple CSC intended for unit testing.

    Supported commands:

    * ``setScalars`` and ``setArrays``: output the provided data using the
      corresponding event and telemetry topics. Note that this violates
      the convention that telemetry is output at regular intervals,
      but it makes unit tests much easier to write.
    * ``wait``: wait for the specified amount of time, and, if requested,
      raise an exception. One use for this is to test command timeout
      by specifying a long wait and waiting a shorter time for the command
      to finish. Another use is to test multiple simultaneous commands,
      since ``wait`` supports this.
    * The standard state transition commands do the usual thing
      and output the ``summaryState`` event. The ``exitControl``
      command shuts the CSC down.

    Parameters
    ----------
    index : `int`
        Index of Test component; each unit test method
        should use a different index.
    config_dir : `str`, optional
        Path to configuration files.
    initial_state : `salobj.State`, optional
        The initial state of the CSC. Typically one of:

        * `salobj.State.ENABLED` if you want the CSC immediately usable.
        * `salobj.State.STANDBY` if you want full emulation of a CSC.
    settings_to_apply : `str`, optional
        Settings to apply if ``initial_state`` is `State.DISABLED`
        or `State.ENABLED`.
    simulation_mode : `int`, optional
        Simulation mode. The only allowed value is 0.
    schema_path : `str`, `pathlib.Path` or `None`, optional
        Path to a schema file used to validate configuration files.
        This is only for testing the deprecated ``schema_path``
        `ConfigurableCsc` constructor argument;
        real CSCs should _not_ provide this argument.
        If not None then `ConfigurableCsc` is called with this,
        instead of the ``config_schema`` argument.


    Raises
    ------
    ValueError
        If ``config_dir`` is not a directory or ``initial_state`` is invalid.
    salobj.ExpectedError
        If ``simulation_mode`` is invalid.
        Note: you will only see this error if you await `start_task`.

    Notes
    -----
    Unlike a normal CSC this one does not output telemetry at regular
    intervals. Instead, in order to simplify unit tests, it outputs
    the ``arrays`` and ``scalars`` telemetry topics in reponse to
    the ``setArrays`` or ``setScalars`` command (just like the ``arrays``
    and ``scalars`` events). That makes it more predictable when this
    data will appear. Note that the ``heartbeat`` event is output at
    regular intervals, as for any CSC.

    Also, unlike most normal configurable CSCs, this one does not need to be
    configured in order to be used (though self.config will be None).
    Thus it is safe to start this CSC in the `salobj.State.ENABLED` state.

    **Error Codes**

    * 1: the fault command was executed
    """

    enable_cmdline_state = True
    valid_simulation_modes = [0]
    version = __version__
    __test__ = False  # stop pytest from warning that this is not a test

    def __init__(
        self,
        index,
        config_dir=None,
        initial_state=State.STANDBY,
        settings_to_apply="",
        simulation_mode=0,
        schema_path=None,
    ):
        if schema_path is not None:
            config_kwargs = dict(schema_path=schema_path)
        else:
            config_kwargs = dict(config_schema=CONFIG_SCHEMA)
        print("config_kwargs=", config_kwargs)
        super().__init__(
            "Test",
            config_dir=config_dir,
            index=index,
            initial_state=initial_state,
            settings_to_apply=settings_to_apply,
            simulation_mode=simulation_mode,
            **config_kwargs,
        )
        self.cmd_wait.allow_multiple_callbacks = True
        self.config = None

    def do_setArrays(self, data):
        """Execute the setArrays command."""
        self.assert_enabled()
        self.log.info("executing setArrays")
        data_dict = self.as_dict(data, self.arrays_fields)
        self.evt_arrays.set(**data_dict)
        self.tel_arrays.set(**data_dict)
        self.evt_arrays.put()
        self.tel_arrays.put()

    def do_setScalars(self, data):
        """Execute the setScalars command."""
        self.assert_enabled()
        self.log.info("executing setScalars")
        data_dict = self.as_dict(data, self.scalars_fields)
        self.evt_scalars.set(**data_dict)
        self.tel_scalars.set(**data_dict)
        self.evt_scalars.put()
        self.tel_scalars.put()

    def as_dict(self, data, fields):
        """Return the specified fields from a data struct as a dict.

        Parameters
        ----------
        data : `any`
            The data to copy.
        fields : `list` [`str`]
            The names of the fields of ``data`` to copy.
        """
        ret = dict()
        for field in fields:
            ret[field] = getattr(data, field)
        return ret

    def do_fault(self, data):
        """Execute the fault command.

        Change the summary state to State.FAULT
        """
        self.log.warning("executing the fault command")
        self.fault(code=1, report="executing the fault command")

    async def do_wait(self, data):
        """Execute the wait command by waiting for the specified duration.

        If duration is negative then wait for abs(duration) but do not
        acknowledge the command as "in progress". This is useful for
        testing command timeout.
        """
        self.assert_enabled()
        if data.duration >= 0:
            self.cmd_wait.ack_in_progress(data, timeout=data.duration)
        await asyncio.sleep(abs(data.duration))

    @property
    def field_type(self):
        """Get a dict of field_name: element type."""
        return dict(
            boolean0=bool,
            byte0=np.uint8,
            char0=str,
            short0=np.int16,
            int0=np.int32,
            long0=np.int32,
            longLong0=np.int64,
            octet0=np.uint8,
            unsignedShort0=np.uint16,
            unsignedInt0=np.uint32,
            unsignedLong0=np.uint32,
            float0=np.single,
            double0=np.double,
            string0=str,
        )

    @property
    def arrays_fields(self):
        """Get a tuple of the fields in an arrays struct."""
        return (
            "boolean0",
            "byte0",
            "short0",
            "int0",
            "long0",
            "longLong0",
            "octet0",
            "unsignedShort0",
            "unsignedInt0",
            "unsignedLong0",
            "float0",
            "double0",
        )

    @property
    def scalars_fields(self):
        """Get a tuple of the fields in a scalars struct."""
        return (
            "boolean0",
            "byte0",
            "char0",
            "short0",
            "int0",
            "long0",
            "longLong0",
            "octet0",
            "unsignedShort0",
            "unsignedInt0",
            "unsignedLong0",
            "float0",
            "double0",
            "string0",
        )

    @property
    def int_fields(self):
        """Get a tuple of the integer fields in a struct."""
        return (
            "byte0",
            "short0",
            "int0",
            "long0",
            "longLong0",
            "octet0",
            "unsignedShort0",
            "unsignedInt0",
            "unsignedLong0",
        )

    def assert_arrays_equal(self, arrays1, arrays2):
        """Assert that two arrays data structs are equal.

        The types need not match; each struct can be command, event
        or telemetry data.
        """
        # use reversed so boolean0 is not compared first,
        # as a discrepancy there is harder to interpret
        for field in reversed(self.arrays_fields):
            field_arr1 = getattr(arrays1, field)
            field_arr2 = getattr(arrays2, field)
            if not np.array_equal(field_arr1, field_arr2):
                raise AssertionError(
                    f"arrays1.{field} = {field_arr1} != {field_arr2} = arrays2.{field}"
                )

    def assert_scalars_equal(self, scalars1, scalars2):
        """Assert that two scalars data structs are equal.

        The types need not match; each struct can be command, event
        or telemetry data.
        """
        # use reversed so boolean0 is not compared first,
        # as a discrepancy there is harder to interpret
        for field in reversed(self.scalars_fields):
            field_val1 = getattr(scalars1, field)
            field_val2 = getattr(scalars2, field)
            if field_val1 != field_val2:
                raise AssertionError(
                    f"scalars1.{field} = {field_val1} != {field_val2} = scalars2.{field}"
                )

    def make_random_cmd_arrays(self):
        return self._make_random_arrays(dtype=self.cmd_setArrays.DataType)

    def make_random_evt_arrays(self):
        return self._make_random_arrays(dtype=self.evt_arrays.DataType)

    def make_random_tel_arrays(self):
        return self._make_random_arrays(dtype=self.tel_arrays.DataType)

    def make_random_cmd_scalars(self):
        return self._make_random_scalars(dtype=self.cmd_setScalars.DataType)

    def make_random_evt_scalars(self):
        return self._make_random_scalars(dtype=self.evt_scalars.DataType)

    def make_random_tel_scalars(self):
        return self._make_random_scalars(dtype=self.tel_scalars.DataType)

    def _make_random_arrays(self, dtype):
        """Make random arrays data using numpy.random."""
        data = dtype()
        nelts = len(data.boolean0)
        data.boolean0 = np.random.choice([False, True], size=(nelts,))
        for field in self.int_fields:
            field_type = self.field_type[field]
            nelts = len(getattr(data, field))
            iinfo = np.iinfo(field_type)
            field_data = np.random.randint(
                iinfo.min, iinfo.max, size=(nelts,), dtype=field_type
            )
            setattr(data, field, field_data)
        for field in ("float0", "double0"):
            field_type = self.field_type[field]
            nelts = len(getattr(data, field))
            field_data = np.array(
                np.random.uniform(-1e5, 1e5, size=(nelts,)), dtype=field_type
            )
            setattr(data, field, field_data)
        return data

    def _make_random_scalars(self, dtype):
        """Make random data for cmd_setScalars using numpy.random."""
        data = dtype()
        data.boolean0 = np.random.choice([False, True])
        printable_chars = [c for c in string.printable]
        data.char0 = "".join(np.random.choice(printable_chars, size=(20,)))
        data.string0 = "".join(np.random.choice(printable_chars, size=(20,)))
        for field in self.int_fields:
            field_type = self.field_type[field]
            iinfo = np.iinfo(field_type)
            field_data = np.random.randint(iinfo.min, iinfo.max, dtype=field_type)
            setattr(data, field, field_data)
        for field in ("float0", "double0"):
            field_type = self.field_type[field]
            field_data = field_type(np.random.uniform(-1e5, 1e5))
            setattr(data, field, field_data)

        data.float0 = np.array(np.random.uniform(-1e5, 1e5), dtype=np.single)
        data.double0 = np.random.uniform(-1e5, 1e5)
        return data

    @staticmethod
    def get_config_pkg():
        return "ts_config_ocs"

    async def configure(self, config):
        self.config = config



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import io
import os
import unittest

import astropy.time
import moto

from lsst.ts import salobj


class AsyncS3BucketTest(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.bucket_name = "async_bucket_test"
        self.file_data = b"Data for the test case"
        self.key = "test_file"
        self.bucket = salobj.AsyncS3Bucket(self.bucket_name, create=True, domock=True)
        self.fileobj = io.BytesIO(self.file_data)

    def tearDown(self):
        self.bucket.stop_mock()

    async def test_attributes(self):
        self.assertEqual(self.bucket.name, self.bucket_name)

    async def test_blank_s3_endpoint_url(self):
        with salobj.modify_environ(S3_ENDPOINT_URL=""):
            bucket = salobj.AsyncS3Bucket(self.bucket_name)
            self.assertIn(
                "amazon", bucket.service_resource.meta.client.meta.endpoint_url
            )

    async def test_no_s3_endpoint_url(self):
        # Clear "S3_ENDPOINT_URL" if it exists.
        with salobj.modify_environ(S3_ENDPOINT_URL=None):
            bucket = salobj.AsyncS3Bucket(self.bucket_name)
            self.assertIn(
                "amazon", bucket.service_resource.meta.client.meta.endpoint_url
            )

    async def test_specified_s3_endpoint_url(self):
        endpoint_url = "http://foo.bar.edu:9000"
        with salobj.modify_environ(S3_ENDPOINT_URL=endpoint_url):
            bucket = salobj.AsyncS3Bucket(self.bucket_name)
            self.assertEqual(
                bucket.service_resource.meta.client.meta.endpoint_url, endpoint_url
            )

    async def test_file_transfer(self):
        await self.bucket.upload(fileobj=self.fileobj, key=self.key)
        roundtrip_fileobj = await self.bucket.download(key=self.key)
        roundtrip_data = roundtrip_fileobj.read()
        self.assertEqual(self.file_data, roundtrip_data)

    async def test_exists(self):
        should_be_false = await self.bucket.exists(key="no_such_file")
        self.assertFalse(should_be_false)

        await self.bucket.upload(fileobj=self.fileobj, key=self.key)
        should_be_true = await self.bucket.exists(key=self.key)
        self.assertTrue(should_be_true)

    async def test_size(self):
        await self.bucket.upload(fileobj=self.fileobj, key=self.key)
        reported_size = await self.bucket.size(key=self.key)
        self.assertEqual(reported_size, len(self.file_data))

    async def test_callbacks(self):
        """Test callback functions with file transfers."""
        uploaded_nbytes = []
        downloaded_nbytes = []

        def upload_callback(nbytes):
            nonlocal uploaded_nbytes
            uploaded_nbytes.append(nbytes)

        def download_callback(nbytes):
            nonlocal downloaded_nbytes
            downloaded_nbytes.append(nbytes)

        await self.bucket.upload(
            fileobj=self.fileobj, key=self.key, callback=upload_callback
        )
        roundtrip_fileobj = await self.bucket.download(
            key=self.key, callback=download_callback
        )
        roundtrip_data = roundtrip_fileobj.getbuffer()
        self.assertEqual(self.file_data, roundtrip_data)
        self.assertGreaterEqual(len(uploaded_nbytes), 1)
        self.assertGreaterEqual(len(downloaded_nbytes), 1)
        self.assertEqual(sum(uploaded_nbytes), len(self.file_data))
        self.assertEqual(sum(downloaded_nbytes), len(self.file_data))


class AsyncS3BucketClassmethodTest(unittest.IsolatedAsyncioTestCase):
    async def test_make_bucket_name_good(self):
        s3instance = "5TEST"
        expected_name = "rubinobs-lfa-5test"
        name = salobj.AsyncS3Bucket.make_bucket_name(s3instance=s3instance)
        self.assertEqual(name, expected_name)

        expected_name = "rubinobs-lfa-5test"
        name = salobj.AsyncS3Bucket.make_bucket_name(s3instance=s3instance)
        self.assertEqual(name, expected_name)

        s3category = "Other3"
        expected_name = "rubinobs-other3-5test"
        name = salobj.AsyncS3Bucket.make_bucket_name(
            s3instance=s3instance,
            s3category=s3category,
        )
        self.assertEqual(name, expected_name)

    async def test_make_bucket_name_bad(self):
        good_kwargs = dict(s3instance="TEST", s3category="other")
        expected_name = "rubinobs-other-test"
        name = salobj.AsyncS3Bucket.make_bucket_name(**good_kwargs)
        self.assertEqual(name, expected_name)

        for argname in good_kwargs:
            for badvalue in (
                ".foo",
                "foo.",
                "#foo",
                "fo#o",
                "@foo",
                "f@oo",
                "_foo",
                "fo_o",
            ):
                with self.subTest(argname=argname, badvalue=badvalue):
                    bad_kwargs = good_kwargs.copy()
                    bad_kwargs[argname] = badvalue
                    with self.assertRaises(ValueError):
                        salobj.AsyncS3Bucket.make_bucket_name(**bad_kwargs)

    async def test_make_key(self):
        # Try a date such that 12 hours earlier is just barely the previous day
        date = astropy.time.Time("2020-04-02T11:59:59.999", scale="tai")
        salname = "Foo"
        salindexname = "Blue"
        generator = "testFiberSpecBlue"
        key = salobj.AsyncS3Bucket.make_key(
            salname=salname, salindexname=salindexname, generator=generator, date=date
        )
        expected_key = (
            "Foo:Blue/testFiberSpecBlue/2020/04/01/"
            "Foo:Blue_testFiberSpecBlue_2020-04-02T11:59:59.999.dat"
        )
        self.assertEqual(key, expected_key)

        # Repeat the test with a date that rounds up to the next second.
        date = astropy.time.Time("2020-04-02T11:59:59.9999", scale="tai")
        key = salobj.AsyncS3Bucket.make_key(
            salname=salname, salindexname=salindexname, generator=generator, date=date
        )
        expected_key = (
            "Foo:Blue/testFiberSpecBlue/2020/04/02/"
            "Foo:Blue_testFiberSpecBlue_2020-04-02T12:00:00.000.dat"
        )
        self.assertEqual(key, expected_key)

        # Repeat the test with no sal index name
        key = salobj.AsyncS3Bucket.make_key(
            salname=salname, salindexname=None, generator=generator, date=date
        )
        expected_key = (
            "Foo/testFiberSpecBlue/2020/04/02/"
            "Foo_testFiberSpecBlue_2020-04-02T12:00:00.000.dat"
        )
        self.assertEqual(key, expected_key)

        # Repeat the test with an integer sal index name
        key = salobj.AsyncS3Bucket.make_key(
            salname=salname, salindexname=5, generator=generator, date=date
        )
        expected_key = (
            "Foo:5/testFiberSpecBlue/2020/04/02/"
            "Foo:5_testFiberSpecBlue_2020-04-02T12:00:00.000.dat"
        )
        self.assertEqual(key, expected_key)

        # Repeat the test with a specified value for "other"
        key = salobj.AsyncS3Bucket.make_key(
            salname=salname,
            salindexname=5,
            generator=generator,
            date=date,
            other="othertext",
        )
        expected_key = (
            "Foo:5/testFiberSpecBlue/2020/04/02/"
            "Foo:5_testFiberSpecBlue_othertext.dat"
        )
        self.assertEqual(key, expected_key)

        # Repeat the test with a specified value for "suffix"
        key = salobj.AsyncS3Bucket.make_key(
            salname=salname,
            salindexname=5,
            generator=generator,
            date=date,
            suffix="suffixtext",
        )
        expected_key = (
            "Foo:5/testFiberSpecBlue/2020/04/02/"
            "Foo:5_testFiberSpecBlue_2020-04-02T12:00:00.000suffixtext"
        )
        self.assertEqual(key, expected_key)

    def test_env_var_secrets(self):
        """Check that moto.mock ovewrites authorization env vars."""
        env_names = (
            "AWS_ACCESS_KEY_ID",
            "AWS_SECRET_ACCESS_KEY",
        )
        env_dict = {name: f"arbitrary value for {name}" for name in env_names}
        with salobj.modify_environ(**env_dict):
            with moto.mock_s3():
                for name, my_value in env_dict.items():
                    self.assertNotEqual(os.environ[name], my_value)
            for name, my_value in env_dict.items():
                self.assertEqual(os.environ[name], my_value)



#!/usr/bin/env python
# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import asyncio

import SALPY_Test

SAL__CMD_COMPLETE = 303


class MinimalSALPYController:
    def __init__(self, index, initial_log_level):
        print(
            f"SALPYController: starting with index={index}, "
            f"initial_log_level={initial_log_level}"
        )
        self.manager = SALPY_Test.SAL_Test(index)
        self.manager.setDebugLevel(0)
        self.manager.salEventPub("Test_logevent_logLevel")
        self.manager.salTelemetryPub("Test_scalars")
        self.manager.salProcessor("Test_command_setLogLevel")
        self.initial_log_level = initial_log_level

    @classmethod
    def make_from_cmd_line(cls):
        """Make an instance from the command line."""
        parser = argparse.ArgumentParser("Run a minimal SALPY Test controller")
        parser.add_argument("index", type=int, help="Script SAL Component index")
        parser.add_argument("initial_log_level", type=int, help="Initial log level")
        args = parser.parse_args()
        return MinimalSALPYController(
            index=args.index, initial_log_level=args.initial_log_level
        )

    @classmethod
    async def amain(cls):
        """Make and run a controller."""
        controller = cls.make_from_cmd_line()
        await controller.run()

    async def run(self):
        setLogLevel_data = SALPY_Test.Test_command_setLogLevelC()
        logLevel_data = SALPY_Test.Test_logevent_logLevelC()
        scalars_data = SALPY_Test.Test_scalarsC()
        print(
            f"SALPYController: writing initial logLevel.level={self.initial_log_level} event"
        )
        logLevel_data.level = self.initial_log_level
        self.manager.logEvent_logLevel(logLevel_data, 1)

        while True:
            while True:
                cmdid = self.manager.acceptCommand_setLogLevel(setLogLevel_data)
                if cmdid > 0:
                    break
                elif cmdid < 0:
                    raise RuntimeError(
                        f"SALPYController: error reading setLogLevel command; cmdid={cmdid}"
                    )
                await asyncio.sleep(0.1)
            print(
                f"SALPYController: read setLogLevel(cmdid={cmdid}; "
                f"level={setLogLevel_data.level})"
            )
            self.manager.ackCommand_setLogLevel(cmdid, SAL__CMD_COMPLETE, 0, "")
            await asyncio.sleep(0.001)

            print(
                f"SALPYController: writing logLevel={logLevel_data.level} event "
                "and the same value in scalars.int0 telemetry"
            )
            logLevel_data.level = setLogLevel_data.level
            scalars_data.int0 = setLogLevel_data.level
            self.manager.logEvent_logLevel(logLevel_data, 1)
            self.manager.putSample_scalars(scalars_data)
            await asyncio.sleep(0.001)
            if setLogLevel_data.level == 0:
                print("SALPYController: quitting")
                break

        # Give remotes time to read final DDS messages before closing
        # the domain participant.
        await asyncio.sleep(1)
        self.manager.salShutdown()


if __name__ == "__main__":
    asyncio.run(MinimalSALPYController.amain())
    print("SALPYController: done")



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import unittest
import pathlib

from lsst.ts import idl
from lsst.ts import salobj


class IdlParserTestCase(unittest.TestCase):
    def setUp(self):
        self.data_path = pathlib.Path(__file__).resolve().parents[0] / "data"

    def test_parse_simple_with_metadata(self):
        self.check_parse_simple(has_metadata=True)

    def test_parse_simple_without_metadata(self):
        self.check_parse_simple(has_metadata=False)

    def check_parse_simple(self, has_metadata):
        """Parse one of the Simple IDL files and check all of the
        resulting metadata.

        Note that the Simple IDL file is a simplified version of the Test IDL
        file, with only a few topics and removing user-defined fields
        with duplicate types.
        """
        if has_metadata:
            idl_path = self.data_path / "sal_revCoded_SimpleWithMetadata.idl"
        else:
            idl_path = self.data_path / "sal_revCoded_SimpleWithoutMetadata.idl"
        metadata = salobj.parse_idl(name="Simple", idl_path=idl_path)
        self.assertEqual(metadata.name, "Simple")
        self.assertTrue(idl_path.samefile(metadata.idl_path))
        if has_metadata:
            self.assertEqual(metadata.sal_version, "4.1.0")
            self.assertEqual(metadata.xml_version, "1.0.0")
        else:
            self.assertIsNone(metadata.sal_version)
            self.assertIsNone(metadata.xml_version)
        self.assertEqual(
            set(metadata.topic_info.keys()),
            set(("command_setArrays", "logevent_scalars")),
        )

        # Dict of field name: expected type name
        field_types = dict(
            TestID="long",
            private_revCode="string",
            private_sndStamp="double",
            private_rcvStamp="double",
            private_seqNum="long",
            private_identity="string",
            private_origin="long",
            boolean0="boolean",
            char0="string",
            short0="short",
            long0="long",
            longLong0="long long",
            octet0="octet",
            string0="string",
            unsignedShort0="unsigned short",
            unsignedLong0="unsigned long",
            float0="float",
            double0="double",
            priority="long",
        )

        for sal_topic_name, topic_metadata in metadata.topic_info.items():
            with self.subTest(sal_topic_name=sal_topic_name):
                self.assertEqual(topic_metadata.sal_name, sal_topic_name)
                self.assertEqual(topic_metadata.version_hash, "1234abcd")
                if has_metadata:
                    expected_description = f"Description of {sal_topic_name}"
                    self.assertEqual(topic_metadata.description, expected_description)
                else:
                    self.assertIsNone(topic_metadata.description)
                expected_field_names = set(field_types.keys())
                if sal_topic_name == "command_setArrays":
                    expected_field_names = set(
                        name
                        for name in expected_field_names
                        if name not in ("char0", "string0", "priority")
                    )

                self.assertEqual(
                    set(topic_metadata.field_info.keys()), expected_field_names
                )
                for field_name, field_metadata in topic_metadata.field_info.items():
                    self.assertEqual(field_metadata.name, field_name)
                    if has_metadata:
                        expected_units = (
                            "secs"
                            if field_metadata.name.endswith("Stamp")
                            else "unitless"
                        )
                        self.assertEqual(field_metadata.units, expected_units)
                        self.assertEqual(
                            field_metadata.description, f"Description of {field_name}"
                        )
                    else:
                        self.assertIsNone(field_metadata.units)
                        self.assertIsNone(field_metadata.description)
                    expected_str_length = {"string0": 20, "private_revCode": 8}.get(
                        field_name
                    )
                    self.assertEqual(field_metadata.str_length, expected_str_length)
                    self.assertEqual(field_metadata.type_name, field_types[field_name])
                    if sal_topic_name == "command_setArrays":
                        expected_array_length = 5 if field_name.endswith("0") else None
                    else:
                        expected_array_length = None
                    self.assertEqual(field_metadata.array_length, expected_array_length)

    def test_test_idl(self):
        """Test a Test IDL file generated by the current version of ts_sal.

        This is to catch format changes relative to the local Simple IDL files.
        It is not as thorough as the tests for the local Simple IDL files
        because I don't want to make too many assumptions about minor details
        of the XML.
        """
        idl_path = idl.get_idl_dir() / "sal_revCoded_Test.idl"
        metadata = salobj.parse_idl(name="Test", idl_path=idl_path)
        self.assertEqual(metadata.name, "Test")
        self.assertTrue(idl_path.samefile(metadata.idl_path))

        # Check that names match between info dict keys
        # and the name in the metadata.
        for sal_topic_name, topic_metadata in metadata.topic_info.items():
            self.assertEqual(topic_metadata.sal_name, sal_topic_name)
            for field_name, field_metadata in topic_metadata.field_info.items():
                self.assertEqual(field_metadata.name, field_name)

        # Check that some of the expected topic names are present.
        some_expected_topic_names = (
            "command_enable",
            "command_setArrays",
            "command_setScalars",
            "logevent_arrays",
            "logevent_scalars",
            "arrays",
            "scalars",
        )
        self.assertTrue(
            set(some_expected_topic_names).issubset(set(metadata.topic_info.keys()))
        )

        # Dict of field name: expected type name
        field_types = dict(
            TestID="long",
            private_revCode="string",
            private_sndStamp="double",
            private_rcvStamp="double",
            private_seqNum="long",
            private_identity="string",
            private_origin="long",
            private_host="long",
            boolean0="boolean",
            byte0="octet",
            char0="string",
            short0="short",
            int0="long",
            long0="long",
            longLong0="long long",
            octet0="octet",
            string0="string",
            unsignedShort0="unsigned short",
            unsignedInt0="unsigned long",
            unsignedLong0="unsigned long",
            float0="float",
            double0="double",
            priority="long",
        )

        # Check some details of arrays topics, including data type,
        # array length and string length.
        for topic_name in ("arrays", "logevent_arrays", "command_setArrays"):
            with self.subTest(topic_name=topic_name):
                topic_metadata = metadata.topic_info[topic_name]
                expected_field_names = set(field_types.keys())
                expected_field_names.remove("string0")  # only in scalars
                if not topic_name.startswith("logevent_"):
                    expected_field_names.remove("priority")
                # There is no char0 field in arrays
                expected_field_names.remove("char0")
                self.assertEqual(
                    set(topic_metadata.field_info.keys()), expected_field_names
                )
                for field_metadata in topic_metadata.field_info.values():
                    if field_metadata.name[-1] != "0":
                        self.assertIsNone(field_metadata.array_length)
                    elif field_metadata.name == "char0":
                        self.assertIsNone(field_metadata.array_length)
                    else:
                        self.assertEqual(field_metadata.array_length, 5)
                    if field_metadata.name == "TestID":
                        self.assertIsNone(field_metadata.description)
                    else:
                        self.assertIsInstance(field_metadata.description, str)
                    self.assertEqual(
                        field_metadata.type_name, field_types[field_metadata.name]
                    )

        # Check some details of scalars topics, including data type,
        # array length and string length.
        for topic_name in ("scalars", "logevent_scalars", "command_setScalars"):
            with self.subTest(topic_name=topic_name):
                topic_metadata = metadata.topic_info[topic_name]
                expected_field_names = set(field_types.keys())
                if not topic_name.startswith("logevent_"):
                    expected_field_names.remove("priority")
                self.assertEqual(
                    set(topic_metadata.field_info.keys()), expected_field_names
                )
                for field_metadata in topic_metadata.field_info.values():
                    self.assertIsNone(field_metadata.array_length)
                    if field_metadata.name == "TestID":
                        self.assertIsNone(field_metadata.description)
                    else:
                        self.assertIsInstance(field_metadata.description, str)
                    self.assertEqual(
                        field_metadata.type_name, field_types[field_metadata.name]
                    )


if __name__ == "__main__":
    unittest.main()



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["Controller", "OPTIONAL_COMMAND_NAMES"]

import asyncio

from . import base
from .domain import Domain
from .sal_info import SalInfo
from .topics import ControllerEvent, ControllerTelemetry, ControllerCommand
from .sal_log_handler import SalLogHandler

# Set of generic commands that need not have an associated do_ method.
# TODO DM-17157: Remove this constant if and when ``SALSubsystems.xml``
# explicitly lists all generic topics used by every SAL component.
OPTIONAL_COMMAND_NAMES = set(("abort", "enterControl", "setValue", "setSimulationMode"))

# Delay before closing the domain participant (seconds).
# This gives remotes time to read final DDS messages before they disappear.
SHUTDOWN_DELAY = 1


def parse_as_prefix_and_set(items_str):
    """Parse a string as an optional +/- prefix and a set of items.

    Parameters
    ----------
    items_str : `str`
        Data string formatted as an optional prefix of "+" or "-"
        followed by a set of comma-separated items (each of which
        must not contain a comma or space).
        Whitespace is ignored after the optional prefix and after each comma.

    Returns
    -------
    prefix_data : `List` [`str`, `set`]
        A tuple of two items:
        * The prefix: "+", "-", or "" (no prefix).
        * The set of items. Each item is a `str` with no whitespace or comma.
    """
    prefix = ""
    if items_str and items_str[0] in ("+", "-"):
        prefix = items_str[0]
        items_str = items_str[1:]
    data_set = set([name.strip() for name in items_str.split(",")])
    data_set -= set([""])
    return prefix, data_set


class Controller:
    """A class that receives commands for a SAL component
    and sends telemetry and events from that component.

    This class provides much of the behavior for `BaseCsc`,
    basically everything except the standard summary states
    and associated transition commands.

    Parameters
    ----------
    name : `str`
        Name of SAL component.
    index : `int` or `None`, optional
        SAL component index, or 0 or None if the component is not indexed.
        A value is required if the component is indexed.
    do_callbacks : `bool`, optional
        Set ``do_<name>`` methods as callbacks for commands?
        If True then there must be exactly one ``do_<name>`` method
        for each command.

    Attributes
    ----------
    log : `logging.Logger`
        A logger.
    salinfo : `SalInfo`
        SAL info.
    cmd_<command_name> : `topics.ControllerCommand`
        Controller command topic. There is one for each command supported by
        the SAL component.
    evt_<event_name> : `topics.ControllerEvent`
        Controller event topic. There is one for each event topic supported by
        the SAL component.
    tel_<telemetry_name> : `topics.ControllerTelemetry`
        Controller telemetry topic. There is one for each telemetry topic
        supported by the SAL component.

    Notes
    -----
    .. _writing_a_controller:

    **Writing a Controller**

    (To write a CSC see :ref:`Writing a CSC<lsst.ts.salobj-writing_a_csc>`,
    instead)

    To write a controller that is not a CSC (one that does not provide
    the standard summary states and associated state transition commands):

    * Inherit from this class.
    * Provide all
      :ref:`Required Logging Attributes<required_logging_attributes>`;
      these are automatically provided to CSCs, but not other controllers.
    * Implement `close_tasks`.

    Here is an example that makes a Test controller and displays
    the topic-related attributes, but has no code to do anything
    useful with those topics (see `TestCsc` for that)::

        include salobj

        # the index is arbitrary, but a remote must use the same index
        # to talk to this controller
        test_controller = salobj.Controller("Test", index=5)
        print(dir(test_controller))

    You should see the following topic-related attributes:

    * Commands, each an instance of `topics.ControllerCommand`:

        * ``cmd_standby``
        * ``cmd_start``
        * ... and so on for all other standard CSC commands
        * ``cmd_setArrays``
        * ``cmd_setScalars``

    * Events, each an instance of `topics.ControllerEvent`:

        * ``evt_appliedSettingsMatchStart``
        * ``evt_errorCode``
        * ... and so on for all other standard CSC events
        * ``evt_arrays``
        * ``evt_scalars``

    * Telemetry, each an instance of `topics.ControllerTelemetry`
      (note that there are no standard CSC telemetry topics):

        * ``tel_arrays``
        * ``tel_scalars``


    .. _required_logging_attributes:

    **Required Logging Attributes**

    Each `Controller` must support the following topics,
    as specified in ts_xml in ``SALGenerics.xml``:

    * setLogLevel command
    * logLevel event
    * logMessage event
    """

    def __init__(self, name, index=None, *, do_callbacks=False):
        # Task that is set done when controller is started.
        # Initialize to None to detect "not fully constructed"
        # in close methods.
        self.start_task = None

        self.isopen = False

        self.start_called = False
        # Task that is set done when the controller is closed
        self.done_task = asyncio.Future()
        self._do_callbacks = do_callbacks

        domain = Domain()
        try:
            self.salinfo = SalInfo(domain=domain, name=name, index=index)
            new_identity = self.salinfo.name_index
            self.salinfo.identity = new_identity
            domain.default_identity = new_identity
            self.log = self.salinfo.log

            if do_callbacks:
                # This must be called after the cmd_ attributes
                # have been added.
                self._assert_do_methods_present()

            for cmd_name in self.salinfo.command_names:
                cmd = ControllerCommand(self.salinfo, cmd_name)
                setattr(self, cmd.attr_name, cmd)

            for evt_name in self.salinfo.event_names:
                evt = ControllerEvent(self.salinfo, evt_name)
                setattr(self, evt.attr_name, evt)

            for tel_name in self.salinfo.telemetry_names:
                tel = ControllerTelemetry(self.salinfo, tel_name)
                setattr(self, tel.attr_name, tel)

            for required_name in ("logMessage", "logLevel", "authList"):
                if not hasattr(self, f"evt_{required_name}"):
                    raise RuntimeError(f"{self!r} has no {required_name} event")

            self._sal_log_handler = SalLogHandler(controller=self)
            self.log.addHandler(self._sal_log_handler)

            # This task is set done when the CSC is fully started.
            # If `start` fails then the task has an exception set
            # and the CSC is not usable.
            self.start_task = asyncio.create_task(self._protected_start())

        except Exception:
            # Note: Domain.basic_close closes all its SalInfo instances.
            domain.basic_close()
            raise
        self.isopen = True

    async def _protected_start(self):
        """Call `start` and handle exceptions."""
        if self.start_called:
            raise RuntimeError("Start already called")
        self.start_called = True
        try:
            await self.start()
        except asyncio.CancelledError:
            self.log.warning("start canceled")
        except Exception as e:
            self.log.exception("start failed")
            await self.close(exception=e, cancel_start=False)
            raise

    async def start(self):
        """Finish construction."""

        # Allow each remote constructor to begin running its start method.
        await asyncio.sleep(0)

        # Wait for all remote salinfos to start.
        start_tasks = []
        for salinfo in self.domain.salinfo_set:
            if not salinfo.start_called:
                # This is either self.salinfo or (very unusual) a remote
                # constructed with start=False.
                continue
            start_tasks.append(salinfo.start_task)
        await asyncio.gather(*start_tasks)
        await self.salinfo.start()

        # Assign command callbacks; give up if this fails, since the CSC
        # will be at least partly broken.
        if self._do_callbacks:
            try:
                self._assign_cmd_callbacks()
            except Exception:
                self.log.exception(
                    "Failed in start on _assign_cmd_callbacks; quitting."
                )
                await self.close()
                return

        self.put_log_level()
        self.evt_authList.set_put(
            authorizedUsers=", ".join(sorted(self.salinfo.authorized_users)),
            nonAuthorizedCSCs=", ".join(sorted(self.salinfo.non_authorized_cscs)),
        )

    @property
    def domain(self):
        return self.salinfo.domain

    async def close(self, exception=None, cancel_start=True):
        """Shut down, clean up resources and set done_task done.

        May be called multiple times. The first call closes the Controller;
        subsequent calls wait until the Controller is closed.

        Subclasses should override `close_tasks` instead of `close`,
        unless you have a good reason to do otherwise.

        Parameters
        ----------
        exception : `Exception`, optional
            The exception that caused stopping, if any, in which case
            the ``self.done_task`` exception is set to this value.
            Specify `None` for a normal exit, in which case
            the ``self.done_task`` result is set to `None`.
        cancel_start : `bool`, optional
            Cancel the start task? Leave this true unless calling
            this from the start task.

        Notes
        -----
        Removes the SAL log handler, calls `close_tasks` to stop
        all background tasks, pauses briefly to allow final SAL messages
        to be sent, then closes the dds domain.
        """
        if self.start_task is None:
            # Not fully constructed; nothing to do.
            return

        if cancel_start and not self.start_task.done():
            self.start_task.cancel()

        if not self.isopen:
            # Closed or closing (we know this instance is fully constructed
            # because we checked that start_task is not None).
            # Wait for done_task to be finished,
            # ignoring any exception. If you want to know about the exception
            # you can examine done_task yourself.
            try:
                self.log.info("Already closing; waiting for done_task")
                await self.done_task
            except Exception:
                pass
            return

        self.isopen = False
        try:
            await self.close_tasks()
        except Exception:
            self.log.exception("Controller.close_tasks failed; close continues")
        try:
            # Give remotes time to read final DDS messages before closing
            # the domain participant.
            self.log.removeHandler(self._sal_log_handler)
            await asyncio.sleep(SHUTDOWN_DELAY)
            await self.domain.close()
        except asyncio.CancelledError:
            self.domain.basic_close()
        except Exception:
            self.domain.basic_close()
            self.log.exception("Controller.close failed near the end; close continues")
        finally:
            if not self.done_task.done():
                if exception:
                    self.done_task.set_exception(exception)
                else:
                    self.done_task.set_result(None)

    async def close_tasks(self):
        """Shut down pending tasks. Called by `close`.

        Perform all cleanup other than disabling logging to SAL
        and closing the dds domain.
        """
        pass

    def do_setAuthList(self, data):
        """Update the authorization list.

        Parameters
        ----------
        data : ``cmd_setAuthList.DataType``
            Authorization lists.

        Notes
        -----
        Add items if the data string starts with "+", ignoring duplicates
        (both with respect to the existing items and within the data string).
        Remove items if the data string starts with "-", ignoring missing
        items (items specified for removal that do not exist).
        Ignore whitespace after each comma and after the +/- prefix.
        """
        users_prefix, users_set = parse_as_prefix_and_set(data.authorizedUsers)
        # Remove "me" from users list.
        users_set.discard(self.salinfo.domain.user_host)

        cscs_prefix, cscs_set = parse_as_prefix_and_set(data.nonAuthorizedCSCs)
        # Strip :0 suffix, if present, for consistency.
        cscs_set = {csc[:-2] if csc.endswith(":0") else csc for csc in cscs_set}

        if users_prefix == "+":
            # + prefix: add users.
            self.salinfo.authorized_users |= users_set
        elif users_prefix == "-":
            # - prefix: remove users.
            self.salinfo.authorized_users -= users_set
        else:
            # No prefix: replace users.
            self.salinfo.authorized_users = users_set

        if cscs_prefix == "+":
            # + prefix: add CSCs.
            self.salinfo.non_authorized_cscs |= cscs_set
        elif cscs_prefix == "-":
            # - prefix: remove CSCs.
            self.salinfo.non_authorized_cscs -= cscs_set
        else:
            # No prefix: replace CSCs.
            self.salinfo.non_authorized_cscs = cscs_set

        self.evt_authList.set_put(
            authorizedUsers=", ".join(sorted(self.salinfo.authorized_users)),
            nonAuthorizedCSCs=", ".join(sorted(self.salinfo.non_authorized_cscs)),
            force_output=True,
        )

    def do_setLogLevel(self, data):
        """Set logging level.

        Parameters
        ----------
        data : ``cmd_setLogLevel.DataType``
            Logging level.
        """
        self.log.setLevel(data.level)
        self.put_log_level()

    def put_log_level(self):
        """Output the logLevel event."""
        self.evt_logLevel.set_put(level=self.log.getEffectiveLevel(), force_output=True)

    def _assert_do_methods_present(self):
        """Assert that all needed do_<name> methods are present."""
        command_names = self.salinfo.command_names
        do_names = [name for name in dir(self) if name.startswith("do_")]
        supported_command_names = [name[3:] for name in do_names]
        if set(command_names) != set(supported_command_names):
            err_msgs = []
            unsupported_commands = (
                set(command_names)
                - set(supported_command_names)
                - OPTIONAL_COMMAND_NAMES
            )
            if unsupported_commands:
                needed_do_str = ", ".join(
                    f"do_{name}" for name in sorted(unsupported_commands)
                )
                err_msgs.append(f"must add {needed_do_str} methods")
            extra_commands = sorted(set(supported_command_names) - set(command_names))
            if extra_commands:
                extra_do_str = ", ".join(
                    f"do_{name}" for name in sorted(extra_commands)
                )
                err_msgs.append(f"must remove {extra_do_str} methods")
            if not err_msgs:
                return
            err_msg = " and ".join(err_msgs)
            raise TypeError(f"This class {err_msg}")

    def _assign_cmd_callbacks(self):
        """Assign each do_ method as a callback to the appropriate command.

        Must not be called until the command attributes have been added.
        """
        command_names = self.salinfo.command_names
        for cmd_name in command_names:
            cmd = getattr(self, f"cmd_{cmd_name}")
            do_method_name = f"do_{cmd_name}"
            func = getattr(self, do_method_name, None)
            if func is not None:
                cmd.callback = func
            elif cmd_name not in OPTIONAL_COMMAND_NAMES:
                raise RuntimeError(f"Can't find method {do_method_name}")
            else:

                def reject_command(data):
                    raise base.ExpectedError("Not supported by this CSC")

                cmd.callback = reject_command

    async def __aenter__(self):
        await self.start_task
        return self

    async def __aexit__(self, type, value, traceback):
        await self.close()



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import logging
import os
import pathlib
import unittest

import numpy as np

from lsst.ts import salobj

# Long enough to perform any reasonable operation
# including starting a CSC or loading a script (seconds)
STD_TIMEOUT = 60
# Timeout for when we expect no new data (seconds).
NODATA_TIMEOUT = 0.1

np.random.seed(47)

index_gen = salobj.index_generator()
TEST_DATA_DIR = TEST_CONFIG_DIR = pathlib.Path(__file__).resolve().parent / "data"
TEST_CONFIG_DIR = TEST_DATA_DIR / "config"


class FailedCallbackCsc(salobj.TestCsc):
    """A CSC whose do_wait command raises a RuntimeError"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.exc_msg = "do_wait raised an exception on purpose"

    async def do_wait(self, data):
        raise RuntimeError(self.exc_msg)


class ControllerLoggingTestCase(
    salobj.BaseCscTestCase, unittest.IsolatedAsyncioTestCase
):
    def basic_make_csc(self, initial_state, config_dir, simulation_mode):
        return FailedCallbackCsc(
            initial_state=initial_state,
            index=self.next_index(),
            config_dir=config_dir,
            simulation_mode=simulation_mode,
        )

    async def test_logging(self):
        async with self.make_csc(
            initial_state=salobj.State.ENABLED, config_dir=TEST_CONFIG_DIR
        ):
            logLevel = await self.remote.evt_logLevel.next(
                flush=False, timeout=STD_TIMEOUT
            )
            self.assertEqual(logLevel.level, logging.INFO)

            self.remote.evt_logMessage.flush()

            # We may still get one or two startup log messages
            # so read until we see the one we want.
            info_message = "test info message"
            self.csc.log.info(info_message)
            while True:
                msg = await self.remote.evt_logMessage.next(
                    flush=False, timeout=STD_TIMEOUT
                )
                if msg.message == info_message:
                    break
            self.assertEqual(msg.level, logging.INFO)
            self.assertEqual(msg.traceback, "")

            filepath = pathlib.Path(__file__)
            subpath = "/".join(filepath.parts[-2:])
            self.assertTrue(
                msg.filePath.endswith(subpath),
                f"{msg.filePath} does not end with {subpath!r}",
            )
            self.assertEqual(msg.functionName, "test_logging")
            self.assertGreater(msg.lineNumber, 0)
            self.assertEqual(msg.process, os.getpid())

            # Test a warning with an unencodable character
            encodable_message = "test warn message"
            warn_message = encodable_message + "\u2013"
            self.csc.log.warning(warn_message)
            msg = await self.remote.evt_logMessage.next(
                flush=False, timeout=STD_TIMEOUT
            )
            encodable_len = len(encodable_message)
            self.assertEqual(msg.message[0:encodable_len], encodable_message)
            self.assertEqual(msg.level, logging.WARNING)
            self.assertEqual(msg.traceback, "")

            with self.assertRaises(asyncio.TimeoutError):
                await self.remote.evt_logMessage.next(
                    flush=False, timeout=NODATA_TIMEOUT
                )

            await self.remote.cmd_setLogLevel.set_start(
                level=logging.ERROR, timeout=STD_TIMEOUT
            )

            logLevel = await self.remote.evt_logLevel.next(
                flush=False, timeout=STD_TIMEOUT
            )
            self.assertEqual(logLevel.level, logging.ERROR)

            info_message = "test info message"
            self.csc.log.info(info_message)
            with self.assertRaises(asyncio.TimeoutError):
                await self.remote.evt_logMessage.next(
                    flush=False, timeout=NODATA_TIMEOUT
                )

            warn_message = "test warn message"
            self.csc.log.warning(warn_message)
            with self.assertRaises(asyncio.TimeoutError):
                await self.remote.evt_logMessage.next(
                    flush=False, timeout=NODATA_TIMEOUT
                )

            with salobj.assertRaisesAckError():
                await self.remote.cmd_wait.set_start(duration=5, timeout=STD_TIMEOUT)

            msg = await self.remote.evt_logMessage.next(
                flush=False, timeout=STD_TIMEOUT
            )
            self.assertIn(self.csc.exc_msg, msg.traceback)
            self.assertIn("Traceback", msg.traceback)
            self.assertIn("RuntimeError", msg.traceback)
            self.assertEqual(msg.level, logging.ERROR)
            self.assertTrue(msg.filePath.endswith("topics/controller_command.py"))
            self.assertNotEqual(msg.functionName, "")
            self.assertGreater(msg.lineNumber, 0)
            self.assertEqual(msg.process, os.getpid())


if __name__ == "__main__":
    unittest.main()



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["SalLogHandler"]

import logging
import sys


class SalLogHandler(logging.Handler):
    """Log handler that outputs an event topic.

    Parameters
    ----------
    controller : `Controller`
        Controller with
        :ref:`Required Logger Attribute<required_logging_attributes>`
        ``evt_logEvent``.
    """

    def __init__(self, controller):
        self.controller = controller
        super().__init__()

    def emit(self, record):
        try:
            self.format(record)
            if record.exc_text is not None:
                traceback = record.exc_text.encode("utf-8", "replace")
            else:
                traceback = ""
            self.controller.evt_logMessage.set_put(
                name=record.name,
                level=record.levelno,
                message=record.message.encode("utf-8", "replace"),
                traceback=traceback,
                filePath=record.pathname,
                functionName=record.funcName,
                lineNumber=record.lineno,
                process=record.process,
                force_output=True,
            )
        except Exception as e:
            print(f"SalLogHandler.emit failed: {e}", file=sys.stderr)
        finally:
            # The Python formatter documentation suggests clearing ``exc_text``
            # after calling ``format`` to avoid problems with
            # multiple formatters that have different exception formats.
            record.exc_text = ""



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License

__all__ = ["CONFIG_SCHEMA"]

import yaml


# Config schema for TestCSC
CONFIG_SCHEMA = yaml.safe_load(
    """
$schema: http://json-schema.org/draft-07/schema#
$id: https://github.com/lsst-ts/ts_salobj/blob/master/python/lsst/ts/salobj/config_schema.py
# title must end with one or more spaces followed by the schema version, which must begin with "v"
title: Test v1
description: Configuration for the TestCsc
type: object
properties:
  string0:
    type: string
    default: default value for string0
  bool0:
    type: boolean
    default: true
  int0:
    type: integer
    default: 5
  float0:
    type: number
    default: 3.14
  intarr0:
    type: array
    default: [-1, 1]
    items:
      type: integer
  multi_type:
    anyOf:
      - type: integer
        minimum: 1
      - type: string
      - type: "null"
    default: null

required: [string0, bool0, int0, float0, intarr0, multi_type]
additionalProperties: false
"""
)



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["RemoteTelemetry"]

import warnings

from . import read_topic


class RemoteTelemetry(read_topic.ReadTopic):
    """Read a specific telemetry topic.

    Parameters
    ----------
    salinfo : `.SalInfo`
        SAL component information
    name : `str`
        Telemetry topic name
    max_history : `int`, optional
        Deprecated because historical telemetry data is no longer available.
        Must be 0 (or None, but please don't do that) if specified.
    queue_len : `int`, optional
        Number of elements that can be queued for `get_oldest`.
    """

    def __init__(
        self, salinfo, name, max_history=None, queue_len=read_topic.DEFAULT_QUEUE_LEN
    ):
        # TODO DM-26474: remove the max_history argument and this code block.
        if max_history is not None:
            if max_history == 0:
                warnings.warn("max_history is deprecated", DeprecationWarning)
            else:
                raise ValueError(
                    f"max_history={max_history} is deprecated "
                    "and must be 0 (or None, but please don't do that) if specified"
                )

        super().__init__(
            salinfo=salinfo,
            name=name,
            sal_prefix="",
            max_history=0,
            queue_len=queue_len,
        )



#!/usr/bin/env python
# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio

from lsst.ts import salobj


class TopicWriter(salobj.BaseCsc):
    """A simple CSC that writes samples as fast as possible.

    Uses the Test API. Specify settingsToApply:

    * logLevel: to write logLevel events when enabled
    * arrays: to write arrays telemetry when enabled

    Accepts and ignores the fault command (in ANY state),
    to allow testing the speed of commands.
    I suggest not enabling the CSC to test command speed,
    since that will make it busy writing topics.

    Parameters
    ----------
    index : `int`
        SAL index.
    """

    valid_simulation_modes = [0]

    def __init__(self, index):
        print(f"TopicWriter: starting with index={index}")
        super().__init__(name="Test", index=index)
        self.write_task = salobj.make_done_future()
        self.is_log_level = False

    async def do_fault(self, data):
        pass

    async def begin_start(self, data):
        if data.settingsToApply == "logLevel":
            self.is_log_level = True
        elif data.settingsToApply == "arrays":
            self.is_log_level = False
        else:
            raise salobj.ExpectedError(
                f"data.settingsToApply={data.settingsToApply} must be logLevel or arrays"
            )
        print(f"TopicWriter: writing {data.settingsToApply}")

    async def do_setArrays(self, data):
        raise salobj.ExpectedError("Not supported")

    async def do_setScalars(self, data):
        raise salobj.ExpectedError("Not supported")

    async def do_wait(self, data):
        raise salobj.ExpectedError("Not supported")

    async def handle_summary_state(self):
        if self.summary_state == salobj.State.ENABLED:
            if self.write_task.done():
                if self.is_log_level:
                    self.write_task = asyncio.create_task(self.write_log_level())
                else:
                    self.write_task = asyncio.create_task(self.write_arrays())
        else:
            self.write_task.cancel()

    async def write_arrays(self):
        i = 0
        while True:
            self.tel_arrays.data.int0[0] = i
            self.tel_arrays.put()
            await asyncio.sleep(0)
            i += 1

    async def write_log_level(self):
        i = 0
        while True:
            self.evt_logLevel.set_put(level=i)
            await asyncio.sleep(0)
            i += 1


if __name__ == "__main__":
    asyncio.run(TopicWriter.amain(index=True))
    print("TopicWriter: done")



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["BaseTopic"]

import abc

import ddsutil

# dict of sal_prefix: attr_prefix: the prefix used for
# Controller and Remote topic attributes.
_ATTR_PREFIXES = {"": "tel_", "logevent_": "evt_", "command_": "cmd_"}


class BaseTopic(abc.ABC):
    r"""Base class for topics.

    Parameters
    ----------
    salinfo : `.SalInfo`
        SAL component information
    name : `str`
        Topic name, without a "command\_" or "logevent\_" prefix.
    sal_prefix : `str`
        SAL topic prefix: one of "command\_", "logevent\_" or ""

    Raises
    ------
    RuntimeError
        If the topic cannot be constructed.

    Attributes
    ----------
    salinfo : `.SalInfo`
        The ``salinfo`` constructor argument.
    name : `str`
        The ``name`` constructor argument.
    sal_name : `str`
        The topic name used by SAL.
        For example: "logevent_summaryState".
    log : `logging.Logger`
        A logger.
    qos_set : `salobj.QosSet`
        Quality of service set.
    volatile : `bool`
        Is this topic volatile (in which case it has no historical data)?
    attr_name : `str`
        Name of topic attribute in `Controller` and `Remote`.
        For example: "evt_summaryState".
    rev_code : `str`
        Revision hash code for the topic.
        This code changes whenever the schema for the topic changes,
        and it is part of the DDS topic name. For example: "90255bf1"
    dds_name : `str`
        Name of topic seen by DDS.
        For example: "Test_logevent_summaryState_90255bf1".
    """

    def __init__(self, *, salinfo, name, sal_prefix):
        try:
            self.salinfo = salinfo
            self.name = str(name)
            self.sal_name = sal_prefix + self.name
            self.log = salinfo.log.getChild(self.sal_name)

            attr_prefix = "ack_" if name == "ackcmd" else _ATTR_PREFIXES.get(sal_prefix)
            if attr_prefix is None:
                raise ValueError(f"Uknown sal_prefix {sal_prefix!r}")
            self.attr_name = attr_prefix + name

            if name == "ackcmd":
                self.qos_set = salinfo.domain.ackcmd_qos_set
            elif sal_prefix == "command_":
                self.qos_set = salinfo.domain.command_qos_set
            elif sal_prefix == "logevent_":
                self.qos_set = salinfo.domain.event_qos_set
            else:
                self.qos_set = salinfo.domain.telemetry_qos_set

            revname = salinfo.revnames.get(self.sal_name)
            if revname is None:
                raise ValueError(
                    f"Could not find {self.salinfo.name} topic {self.sal_name}"
                )
            self.dds_name = revname.replace("::", "_")
            self.rev_code = self.dds_name.rsplit("_", 1)[1]

            self._type = ddsutil.get_dds_classes_from_idl(
                salinfo.metadata.idl_path, revname
            )
            self._topic = self._type.register_topic(
                salinfo.domain.participant, self.dds_name, self.qos_set.topic_qos
            )

        except Exception as e:
            raise RuntimeError(f"Failed to create topic {salinfo.name}.{name}") from e

    @property
    def DataType(self):
        """The type (class) for a message of this topic.

        When you read or write a message for this topic you are reading
        or writing an instance of `DataType`.

        Notes
        -----
        The preferred way to write a message for a topic is:

        * `RemoteCommand.set_start` to start a command.
        * `CommandEvent.set_put` to write an event.
        * `CommandTelemetry.set_put` to write a telemetry message.

        However, it is also possible to use `DataType` to create a message,
        then write, it as separate operations. For example,
        assuming we have a `Remote` for SAL component "Test"::

            # The preferred way to issue a command:
            await = remote.cmd_wait.set_put(duration=2, timeout=5)

            # But an alternative is to first create the command,
            # then send it, as two separate operations:
            message = remote.cmd_wait.DataType(duration=2)
            await remote.cmd_wait.start(message, timeout=5)

            # Or, even more verbosely:
            message = remote.cmd_wait.DataType()
            message.duration = 2
            await remote.cmd_wait.start(message, timeout=5)
        """
        return self._type.topic_data_class

    @property
    def volatile(self):
        """Does this topic have volatile durability?"""
        return self.qos_set.volatile

    @property
    def metadata(self):
        """Get topic metadata as a `TopicMetadata`, if available,else None."""
        return self.salinfo.metadata.topic_info.get(self.sal_name)

    def __repr__(self):
        return f"{type(self).__name__}({self.salinfo.name}, {self.salinfo.index}, {self.name})"



from .base_topic import *
from .read_topic import *
from .write_topic import *
from .controller_command import *
from .controller_event import *
from .controller_telemetry import *
from .remote_command import *
from .remote_event import *
from .remote_telemetry import *



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = [
    "make_state_transition_dict",
    "get_expected_summary_states",
    "set_summary_state",
]

import asyncio

from .sal_enums import State


def make_state_transition_dict():
    """Make a dict of state transition commands and states

    The keys are (beginning state, ending state).
    The values are a list of tuples:

    * A state transition command
    * The expected state after that command
    """
    ordered_states = (State.OFFLINE, State.STANDBY, State.DISABLED, State.ENABLED)

    basic_state_transition_commands = {
        (State.OFFLINE, State.STANDBY): "enterControl",
        (State.STANDBY, State.DISABLED): "start",
        (State.DISABLED, State.ENABLED): "enable",
        (State.ENABLED, State.DISABLED): "disable",
        (State.DISABLED, State.STANDBY): "standby",
        (State.STANDBY, State.OFFLINE): "exitControl",
    }

    # compute transitions from non-FAULT to all other states
    state_transition_dict = dict()
    for beg_ind, beg_state in enumerate(ordered_states):
        for end_ind, end_state in enumerate(ordered_states):
            if beg_ind == end_ind:
                state_transition_dict[(beg_state, end_state)] = []
                continue
            step = 1 if end_ind > beg_ind else -1
            command_state_list = []
            for next_ind in range(beg_ind, end_ind, step):
                from_state = ordered_states[next_ind]
                to_state = ordered_states[next_ind + step]
                command = basic_state_transition_commands[from_state, to_state]
                command_state_list.append((command, to_state))
            state_transition_dict[(beg_state, end_state)] = command_state_list

    # add transitions from FAULT to all other states
    for end_state in ordered_states:
        command_state_list = [("standby", State.STANDBY)]
        if end_state != State.STANDBY:
            command_state_list += state_transition_dict[State.STANDBY, end_state]
        state_transition_dict[State.FAULT, end_state] = command_state_list
    return state_transition_dict


_STATE_TRANSITION_DICT = make_state_transition_dict()


def get_expected_summary_states(initial_state, final_state):
    """Return all summary states expected when transitioning from
    one state to another.
    """
    cmd_state_list = _STATE_TRANSITION_DICT[(initial_state, final_state)]
    # cmd_state_list lists the state after each state transition
    # command, but we want the initial state, as well.
    return [initial_state] + [cmd_state[1] for cmd_state in cmd_state_list]


async def set_summary_state(remote, state, settingsToApply="", timeout=30):
    """Put a CSC into the specified summary state.

    Parameters
    ----------
    remote : `Remote`
        Remote for the CSC to be enabled.
    state : `State` or `int`
        Desired summary state.
    settingsToApply : `str` or `None`
        SettingsToApply argument for the ``start`` command.
        Ignored unless the CSC has to be taken from state
        `State.STANDBY` to `State.DISABLED`.
    timeout : `float`
        Timeout for each state transition command and a possible initial
        summaryState read (sec).

    Returns
    -------
    states : `list` [`State`]
        A list of the initial summary state and all summary states
        this function transitioned the CSC through,
        ending with the desired state.

    Notes
    -----
    This function assumes the CSC is listening to SAL commands. If the CSC
    is not running then this function will time out (unless the last
    reported summary state has been cached and matches ``state``.
    """
    state = State(state)
    if state == State.FAULT:
        raise ValueError("Cannot go into FAULT state using state transition commands")
    if settingsToApply is None:
        settingsToApply = ""

    # get current summary state
    try:
        state_data = await remote.evt_summaryState.aget(timeout=timeout)
    except asyncio.TimeoutError:
        raise RuntimeError(f"Cannot get summaryState from {remote.salinfo.name}")
    current_state = State(state_data.summaryState)

    states = [current_state]
    if current_state == state:
        # we are already in the desired state
        return states

    command_state_list = _STATE_TRANSITION_DICT[(current_state, state)]

    old_settings_to_apply = remote.cmd_start.data.settingsToApply
    try:
        remote.cmd_start.data.settingsToApply = settingsToApply

        for command, state in command_state_list:
            cmd = getattr(remote, f"cmd_{command}")
            try:
                await cmd.start(timeout=timeout)
            except Exception as e:
                raise RuntimeError(
                    f"Error on cmd=cmd_{command}, initial_state={current_state}: {e}"
                ) from e
            states.append(state)
    finally:
        remote.cmd_start.data.settingsToApply = old_settings_to_apply
    return states



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["AckCmdWriter", "ControllerCommand"]

import asyncio
import inspect
import warnings

from .. import sal_enums
from .. import base
from . import read_topic
from . import write_topic


class AckCmdWriter(write_topic.WriteTopic):
    """Command Acknowledgement writer."""

    def __init__(self, salinfo):
        super().__init__(
            salinfo=salinfo, name="ackcmd", sal_prefix="", min_seq_num=None
        )


class ControllerCommand(read_topic.ReadTopic):
    """Read a specified command topic.

    Parameters
    ----------
    salinfo : `.SalInfo`
        SAL component information
    name : `str`
        Command name
    queue_len : `int`, optional
        Number of elements that can be queued for `get_oldest`.

    Notes
    -----
    Each command must be acknowledged by writing an appropriate ``ackcmd``
    message. If you use a callback function to process the command
    then this happens automatically. Otherwise you must call the `ack` method
    to acknowledge the command yourself, though an initial acknowledgement
    with ``ack=SalRetCode.CMD_ACK`` is always automatically sent when
    the command is read.


    After the initial acknowledgement with ``ack=SalRetCode.CMD_ACK``,
    automatic ackowledgement for callback functions works as follows:

    * If the callback function returns `None` then send a final
      acknowledgement with ``ack=SalRetCode.CMD_COMPLETE``.
    * If the callback function returns an acknowledgement
      (instance of `SalInfo.AckCmdType`) instead of `None`, then send that
      as the final acknowledgement. This is very unusual, but might
      be useful if the callback wants to exit early and leave some
      background task or thread processing the rest of the command.
    * If the callback function raises `asyncio.TimeoutError` then send a
      final acknowledgement with ``ack=SalRetCode.CMD_TIMEOUT``.
    * If the callback function raises `asyncio.CancelledError` then send
      a final acknowledgement with ``ack=SalRetCode.CMD_ABORTED``.
    * If the callback function raises `ExpectedError` then send a final
      acknowledgement with ``ack=SalRetCode.CMD_FAILED`` and
      ``result=f"Failed: {exception}"``.
    * If the callback function raises any other `Exception`
      then do the same as `ExpectedError` and also log a traceback.
    """

    def __init__(self, salinfo, name, queue_len=read_topic.DEFAULT_QUEUE_LEN):
        super().__init__(
            salinfo=salinfo,
            name=name,
            sal_prefix="command_",
            max_history=0,
            queue_len=queue_len,
        )
        # Set false to ignore authorization.
        # The only command should ignore authorization is
        # the command that handles user requests for authorization.
        # TODO DM-26605: Change this to True
        self.authorize = False
        self.cmdtype = salinfo.sal_topic_names.index(self.sal_name)
        if salinfo._ackcmd_writer is None:
            self.salinfo._ackcmd_writer = AckCmdWriter(salinfo=salinfo)

    def ack(self, data, ackcmd):
        """Acknowledge a command by writing a new state.

        Parameters
        ----------
        data : `DataType`
            Data for the command being acknowledged.
        ackcmd : `salobj.AckCmdType`
            Command acknowledgement data.
        """
        self.salinfo._ackcmd_writer.set(
            private_seqNum=data.private_seqNum,
            origin=data.private_origin,
            identity=data.private_identity,
            cmdtype=self.cmdtype,
            ack=ackcmd.ack,
            error=ackcmd.error,
            result=ackcmd.result,
            timeout=ackcmd.timeout,
        )
        self.salinfo._ackcmd_writer.put()

    def ackInProgress(self, data, result=""):
        """Deprecated version of ack_in_progress."""
        # TODO DM-26518: remove this method
        warnings.warn(
            "ackInProgress is deprecated; use ack_in_progress and specify timeout instead",
            DeprecationWarning,
        )
        self.ack_in_progress(data=data, result=result, timeout=0)

    def ack_in_progress(self, data, *, timeout, result=""):
        """Ackowledge this command as "in progress".

        Parameters
        ----------
        data : `DataType`
            Data for the command being acknowledged.
        timeout : `float`
            Estimated command duration (sec).
        result : `str`, optional
            Reason for acknowledgement. Typically left "".
        """
        ackcmd = self.salinfo.make_ackcmd(
            private_seqNum=data.private_seqNum,
            ack=sal_enums.SalRetCode.CMD_INPROGRESS,
            result=result,
            timeout=timeout,
        )
        self.ack(data, ackcmd)

    async def next(self, *, timeout=None):
        """Wait for data, returning old data if found.

        Unlike `RemoteEvent.next` and `RemoteTelemetry.next`,
        the flush argument is not allowed; the only way to flush
        old commands is to call `flush`.

        Parameters
        ----------
        timeout : `float`, optional
            Time limit, in seconds. If None then no time limit.

        Returns
        -------
        data : `DataType`
            Command data.

        Raises
        ------
        RuntimeError
            If a callback function is present.

        Notes
        -----
        Do not modify the data or assume that it will be static.
        If you need a private copy, then copy it yourself.
        """
        return await super().next(flush=False, timeout=timeout)

    def _queue_one_item(self, data):
        """Convert the value to an ``ackcmd`` and queue it."""
        if data.private_seqNum <= 0:
            raise ValueError(f"private_seqNum={data.private_seqNum} must be positive")
        ack = self.salinfo.make_ackcmd(
            private_seqNum=data.private_seqNum, ack=sal_enums.SalRetCode.CMD_ACK
        )
        self.ack(data, ack)
        super()._queue_one_item(data)

    async def _run_callback(self, data):
        """Run the callback function, acknowledge the command,
        and start another wait.

        Parameters
        ----------
        data : `DataType`
            Command data.
        """
        if self.authorize:
            try:
                identity = data.private_identity
                auth_error = None
                if "@" in identity:
                    # A user
                    if (
                        identity != self.salinfo.domain.user_host
                        and identity not in self.salinfo.authorized_users
                    ):
                        auth_error = (
                            f"User identity {identity!r} "
                            f"not self ({self.salinfo.domain.user_host}) "
                            f"and not in {self.salinfo.authorized_users}"
                        )
                elif identity in self.salinfo.non_authorized_cscs:
                    auth_error = f"CSC identity {identity!r} in {self.salinfo.non_authorized_cscs}"
                if auth_error is not None:
                    ack = self.salinfo.make_ackcmd(
                        private_seqNum=data.private_seqNum,
                        ack=sal_enums.SalRetCode.CMD_NOPERM,
                        error=1,
                        result=f"Not authorized: {auth_error}",
                        truncate_result=True,
                    )
                    self.ack(data, ack)
                    return
            except Exception:
                self.log.exception("Error checking identity")

        try:
            result = self._callback(data)
            if inspect.isawaitable(result):
                ack = await result
            else:
                ack = result
            if ack is None:
                ack = self.salinfo.make_ackcmd(
                    private_seqNum=data.private_seqNum,
                    ack=sal_enums.SalRetCode.CMD_COMPLETE,
                    result="Done",
                )
            self.ack(data, ack)
        except asyncio.CancelledError:
            ack = self.salinfo.make_ackcmd(
                private_seqNum=data.private_seqNum,
                ack=sal_enums.SalRetCode.CMD_ABORTED,
                error=1,
                result="Aborted",
                truncate_result=True,
            )
            self.ack(data, ack)
        except asyncio.TimeoutError:
            ack = self.salinfo.make_ackcmd(
                private_seqNum=data.private_seqNum,
                ack=sal_enums.SalRetCode.CMD_TIMEOUT,
                error=1,
                result="Timeout",
                truncate_result=True,
            )
            self.ack(data, ack)
        except Exception as e:
            ack = self.salinfo.make_ackcmd(
                private_seqNum=data.private_seqNum,
                ack=sal_enums.SalRetCode.CMD_FAILED,
                error=1,
                result=f"Failed: {e}",
                truncate_result=True,
            )
            self.ack(data, ack)
            if not isinstance(e, base.ExpectedError):
                self.log.exception(f"Callback {self.callback} failed with data={data}")



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["QueueCapacityChecker", "ReadTopic", "DEFAULT_QUEUE_LEN", "MIN_QUEUE_LEN"]

import asyncio
import bisect
import collections
import inspect
import warnings

import dds

from .base_topic import BaseTopic
from .. import base

# Default value for the ``queue_len`` constructor argument.
DEFAULT_QUEUE_LEN = 100

# Minimum value for the ``queue_len`` constructor argument.
MIN_QUEUE_LEN = 10


class QueueCapacityChecker:
    """Log warnings for a fixed-length queue that should contain
    no more than one item.

    Call `check_nitems` to check the number of items in the queue.
    This will log a warning or error, if appropriate.

    Parameters
    ----------
    descr : `str`
        Brief description of queue, e.g. "python read queue"
        or "DDS read queue".
    log : `logging.Logger`
        Logger to which to write messages.
    queue_len : `int`
        Length of queue

    Attributes
    ----------
    warn_thresholds : `List` [`int`]
        Values for ``warn_threshold`` (see next item). Set to:

        * 5-10 elements, whichever is closest to ``queue_len/10``
        * 1/2 full, but only if ``queue_len >= 20``
        * 9/10 full
        * full

        The corresponding reset thresholds are 1/2 these values.

    warn_threshold : `int` or `None`
        A call to ``check_nitems(n)`` with ``n > warn_threshold``
        will produce a log message and update ``warn_threshold``
        and ``reset__threshold`` as follows:

        *  ``warn_threshold`` is set to the largest warn threshold <= n
           in ``warn_thresholds``, or `None` if the queue is full.
        * ``reset_threshold`` is set to the half of the next lower
          warning threshold.
    reset_threshold : `int` or `None`
        A call to ``check_nitems(n)`` with ``n < reset_threshold``
        will reduce ``warn_threshold`` and ``reset_threshold`` as follows:

        * ``reset_threshold`` is set to the largest reset threshold <= n,
          or `None` if the queue is empty.
        * ``warn_threshold`` is set to the warning threshold
          corresponding to one reset higher reset level.

    Notes
    -----
    Once a message has been logged for a particular threshold,
    no more messages are sent until either the queue fills enough
    to cross the next warning threshold or empties enough to reduce
    the warning threshold.

    Log messages are warnings unless the queue is full. A full queue
    produces an error because data is likely to have been lost.
    """

    def __init__(self, descr, log, queue_len):
        if queue_len < MIN_QUEUE_LEN:
            raise ValueError(
                f"queue_len {queue_len} must be >= MIN_QUEUE_LEN={MIN_QUEUE_LEN}"
            )
        self.descr = descr
        self.log = log
        self.queue_len = queue_len
        warn_thresholds = [
            min(10, max(5, queue_len // 10)),
            queue_len * 9 // 10,
            queue_len,
        ]
        if queue_len >= 20:
            warn_thresholds.append(queue_len // 2)
        self.warn_thresholds = tuple(sorted(warn_thresholds))
        self._reset_thresholds = tuple(
            warn_thresh // 2 for warn_thresh in self.warn_thresholds
        )
        self.warn_threshold = self.warn_thresholds[0]
        self.reset_threshold = None

    def check_nitems(self, nitems):
        """Check the number of items in the queue and log a message
        if appropriate.

        Parameters
        ----------
        nitems : `int`
            Number of elements in the queue.

        Returns
        -------
        did_log: `bool`
            True if a message was logged.
        """
        if self.warn_threshold is not None and nitems >= self.warn_threshold:
            # Issue a warning or error log message
            # and set new warning and reset thresholds.
            if nitems >= self.queue_len:
                self.warn_threshold = None
                self.reset_threshold = self._reset_thresholds[-1]
                self.log.error(
                    f"{self.descr} is full ({self.queue_len} elements); data may be lost"
                )
            else:
                index = bisect.bisect_right(self.warn_thresholds, nitems)
                self.warn_threshold = self.warn_thresholds[index]
                self.reset_threshold = self._reset_thresholds[index - 1]
                self.log.warning(
                    f"{self.descr} is filling: {nitems} of {self.queue_len} elements"
                )
            return True
        elif self.reset_threshold is not None and nitems <= self.reset_threshold:
            # Reset to lower warning and reset thresholds
            index = bisect.bisect_right(self._reset_thresholds, nitems - 1)
            self.warn_threshold = self.warn_thresholds[index]
            if index > 0:
                self.reset_threshold = self._reset_thresholds[index - 1]
            else:
                self.reset_threshold = None
        return False


class ReadTopic(BaseTopic):
    r"""Base class for reading a topic.

    Parameters
    ----------
    salinfo : `.SalInfo`
        SAL component information
    name : `str`
        Topic name, without a "command\_" or "logevent\_" prefix.
    sal_prefix : `str`
        SAL topic prefix: one of "command\_", "logevent\_" or ""
    max_history : `int`
        Maximum number of historical items to read:

        * 0 is required for commands, events, and the ackcmd topic
        * 1 is recommended for telemetry
    queue_len : `int`, optional
        The maximum number of messages that can be read and not dealt with
        by a callback function or `next` before older messages will be dropped.
    filter_ackcmd : `bool`, optional
        Filter out cmdack topics so we only see responses to commands
        that we sent? This is normally what you want, but it is not wanted
        for SAL/Kafka producers.
        Ignored if ``name`` != "ackcmd".

    Raises
    ------
    ValueError
        If max_history < 0.
    ValueError
        If max_history > 0 and the topic is volatile (command or ackcmd).
    ValueError
        If queue_len < MIN_QUEUE_LEN.
    ValueError
        If max_history > queue_len.
    UserWarning
        If max_history > DDS history queue depth or DDS durability service
        history depth for this topic.
        This is a warning rather than an exception, so that the DDS quality
        of service can be changed without breaking existing code.

    Attributes
    ----------
    isopen : `bool`
        Is this read topic open? `True` until `close` is called.
    dds_queue_length_checker : `QueueCapacityChecker`
        Queue length checker for the DDS queue.
    python_queue_length_checker : `QueueCapacityChecker`:
        Queue length checker for the Python queue.

    Notes
    -----
    **Queues**

    There are two queues: a Python queue whose length is set by ``queue_len``
    and a dds queue whose length is set by the DDS Quality of Service file.
    (The Python queue is needed because of limitations in the API for
    the OpenSplice DDS queue, including no access to the most recent message,
    no ability to ask how many messages are on the queue, and no asyncio
    support). In the doc strings for the methods, below, any reference
    to the queue refers to the Python queue.

    Data can be lost from either queue:

    - If this class cannot read messages from the DDS queue fast enough,
      then older messages will be dropped from the DDS queue. You will get
      several warning log messages as the DDS queue fills.
    - As messages are read from the DDS queue they are put on the Python queue.
      If a callback function or `next` does not process data quickly enough
      then older messages are dropped from the Python queue.
      If you have a callback function then you will get several
      warning log messages as the Python queue fills up;
      you get no warning otherwise because `ReadTopic` has no way of knowing
      whether or not you intend to read all messages.

    **Reading**

    Reading is performed by the contained `SalInfo`, which has single
    read loop that reads messages for all topics. This is more efficient
    than having each `ReadTopic` read its own messages.

    **Modifying Messages**

    All functions that return messages return them from some form of internal
    cache. This presents a risk: if any reader modifies a message, then it
    will be modified for all readers of that message. To safely modify a
    returned message, make your own copy with ``copy.copy(data)``.
    """

    def __init__(
        self,
        *,
        salinfo,
        name,
        sal_prefix,
        max_history,
        queue_len=DEFAULT_QUEUE_LEN,
        filter_ackcmd=True,
    ):
        super().__init__(salinfo=salinfo, name=name, sal_prefix=sal_prefix)
        self.isopen = True
        self._allow_multiple_callbacks = False
        if max_history < 0:
            raise ValueError(f"max_history={max_history} must be >= 0")
        if max_history > 0 and self.volatile:
            raise ValueError(f"max_history={max_history} must be 0 for volatile topics")
        if queue_len <= MIN_QUEUE_LEN:
            raise ValueError(
                f"queue_len={queue_len} must be >= MIN_QUEUE_LEN={MIN_QUEUE_LEN}"
            )
        if max_history > queue_len:
            raise ValueError(
                f"max_history={max_history} must be <= queue_len={queue_len}"
            )
        if (
            max_history > self.qos_set.reader_qos.history.depth
            or max_history > self.qos_set.topic_qos.durability_service.history_depth
        ):
            warnings.warn(
                f"max_history={max_history} > history depth={self.qos_set.reader_qos.history.depth} "
                f"and/or {self.qos_set.topic_qos.durability_service.history_depth}; "
                "you will get less historical data than you asked for.",
                UserWarning,
            )
        self._max_history = int(max_history)
        self._data_queue = collections.deque(maxlen=queue_len)
        self._current_data = None
        # Task that `next` waits on.
        # Its result is set to the oldest message on the queue.
        # We do this instead of having `next` itself pop the oldest message
        # because it allows multiple callers of `next` to all get the same
        # message, and it avoids a potential race condition with `flush`.
        self._next_task = base.make_done_future()
        self._callback = None
        self._callback_tasks = set()
        self._callback_loop_task = base.make_done_future()
        self.dds_queue_length_checker = QueueCapacityChecker(
            descr=f"{name} DDS read queue",
            log=self.log,
            queue_len=self.qos_set.reader_qos.history.depth,
        )
        self.python_queue_length_checker = QueueCapacityChecker(
            descr=f"{name} python read queue", log=self.log, queue_len=queue_len
        )
        # Command topics use a different a partition name than
        # all other topics, including ackcmd, and the partition name
        # is part of the publisher and subscriber.
        # This split allows us to create just one subscriber and one publisher
        # for each Controller or Remote:
        # `Controller` only needs a cmd_subscriber and data_publisher,
        # `Remote` only needs a cmd_publisher and data_subscriber.
        if sal_prefix == "command_":
            subscriber = salinfo.cmd_subscriber
        else:
            subscriber = salinfo.data_subscriber
        self._reader = subscriber.create_datareader(
            self._topic, self.qos_set.reader_qos
        )
        # TODO DM-26411: replace ANY_INSTANCE_STATE with ALIVE_INSTANCE_STATE
        # once the OpenSplice issue 00020647 is fixed.
        read_mask = [
            dds.DDSStateKind.NOT_READ_SAMPLE_STATE,
            dds.DDSStateKind.ANY_INSTANCE_STATE,
        ]
        queries = []
        if salinfo.index > 0:
            queries.append(f"{salinfo.name}ID = {salinfo.index}")
        if name == "ackcmd" and filter_ackcmd:
            queries += [
                f"origin = {salinfo.domain.origin}",
                f"identity = '{salinfo.identity}'",
            ]
        if queries:
            full_query = " AND ".join(queries)
            read_condition = dds.QueryCondition(self._reader, read_mask, full_query)
        else:
            read_condition = self._reader.create_readcondition(read_mask)
        self._read_condition = read_condition

        salinfo.add_reader(self)

    @property
    def allow_multiple_callbacks(self):
        """Can callbacks can run simultaneously?

        Notes
        -----
        Ignored for synchronous callbacks because those block
        while running. In particular, if the callback is
        synchronous but launches one or more background jobs
        then the number of those jobs cannot be limited by
        this class.
        """
        return self._allow_multiple_callbacks

    @allow_multiple_callbacks.setter
    def allow_multiple_callbacks(self, allow):
        self._allow_multiple_callbacks = bool(allow)

    @property
    def callback(self):
        """Callback function, or None if there is not one.

        The callback function is called when a new message is received;
        it receives one argument: the message (an object of type
        `topics.BaseTopic.DataType`).

        Raises
        ------
        TypeError
            When setting a new callback if the callback is not None
            and is not callable.

        Notes
        -----
        The callback function can be synchronous or asynchronous
        (e.g. defined with ``async def``).

        Setting a callback flushes the queue, and it will remain empty
        as long as there is a callback.

        `get_oldest` and `next` are prohibited if there is a callback function.
        Technically they could both work, but `get_oldest` would always return
        `None` and `next` would miss messages if they arrived while waiting
        for something else. It seems safer to raise an exception.
        """
        return self._callback

    @callback.setter
    def callback(self, func):
        self._cancel_callbacks()

        if func is None:
            # clear the existing callback
            self._callback = None
            return

        # flush the queue, set a new callback and restart the callback loop
        if not callable(func):
            raise TypeError(f"func={func} not callable")
        self._data_queue.clear()
        self._callback = func
        self._callback_loop_task = asyncio.create_task(self._callback_loop())

    @property
    def has_callback(self):
        """Return True if there is a callback function."""
        return self._callback is not None

    @property
    def has_data(self):
        """Has any data ever been seen for this topic?

        Raises
        ------
        RuntimeError
            If the ``salinfo`` has not started reading.
        """
        self.salinfo.assert_started()
        return self._current_data is not None

    @property
    def nqueued(self):
        """Return the number of messages in the Python queue."""
        return len(self._data_queue)

    @property
    def max_history(self):
        return self._max_history

    def basic_close(self):
        """A synchronous and possibly less thorough version of `close`.

        Intended for exit handlers and constructor error handlers.
        """
        if not self.isopen:
            return
        self.isopen = False
        self._callback = None
        try:
            # These raise RuntimeError if the asyncio loop is not running.
            self._cancel_callbacks()
            self._next_task.cancel()
        except RuntimeError:
            pass
        self._reader.close()
        self._data_queue.clear()

    async def close(self):
        """Shut down and release resources.

        Intended to be called by SalInfo.close(),
        since that tracks all topics.
        """
        self.basic_close()

    async def aget(self, timeout=None):
        """Get the most recent message, or wait for data if no data has
        ever been seen (`has_data` False).

        This method does not change which message will be returned by
        any other method (except for the fact that new data
        will arrive while waiting).

        Parameters
        ----------
        timeout : `float`, optional
            Time limit, in seconds. If None then no time limit.

        Returns
        -------
        data : `DataType`
            The current or next message.

        Raises
        ------
        RuntimeError
            If a callback function is present,
            or if the ``salinfo`` has not started reading.

        Notes
        -----
        Do not modify the returned data. To make a copy that you can
        safely modify, use ``copy.copy(data)``.
        """
        self.salinfo.assert_started()
        if self.has_callback:
            raise RuntimeError("Not allowed because there is a callback function")
        if self._current_data is None:
            if self._next_task.done():
                self._next_task = asyncio.Future()
            await asyncio.wait_for(self._next_task, timeout=timeout)
        return self._current_data

    def flush(self):
        """Flush the queue used by `get_oldest` and `next`.

        This makes `get_oldest` return `None` and `next` wait,
        until a new message arrives.
        It does not change which message will be returned by `aget` or `get`.

        Raises
        ------
        RuntimeError
            If a callback function is present.
        """
        if self.has_callback:
            raise RuntimeError("Not allowed because there is a callback function")
        self._data_queue.clear()

    def get(self, flush=None):
        """Get the most recent message, or `None` if no data has ever been seen
        (`has_data` False).

        This method does not change which message will be returned by `aget`.
        If ``flush=False`` this method also does not modify which message
        will be returned by `get_oldest` and `next`.

        Parameters
        ----------
        flush : `bool`, optional
            Flush the queue? Flushing the queue is deprecated and so
            is specifying this argument.
            False (the default) leaves the cache alone, which has no effect
            on the messages returned by any read method.
            True affects which messages will be returned by `get_oldest`
            and `next`. True has no effect if there is a callback function.
            Note: `None` is treated as `False`, but please do not specify
            it; it only supported for now to handle deprecation warnings..

        Returns
        -------
        data : ``self.DataType`` or `None`
            Return ``self.data`` if data has been read, else `None`.

        Raises
        ------
        RuntimeError
            If the ``salinfo`` has not started reading.
        """
        self.salinfo.assert_started()
        if flush is None:
            flush = False
        else:
            if flush:
                warnings.warn(
                    "flush=True is deprecated for ReadTopic.get",
                    DeprecationWarning,
                )
            else:
                warnings.warn(
                    "Specifying a value for the flush argument is deprecated "
                    "for ReadTopic.get; use the default of False.",
                    DeprecationWarning,
                )

        if flush and not self.has_callback:
            self.flush()
        return self._current_data

    def get_oldest(self):
        """Pop and return the oldest message from the queue, or `None` if the
        queue is empty.

        This is a variant of `next` that does not wait for a new message.
        This method affects which message will be returned by `next`,
        but not which message will be returned by `aget` or `get`.

        Returns
        -------
        data : ``self.DataType`` or `None`
            The oldest message found on the queue, if any, else `None`.

        Raises
        ------
        RuntimeError
            If a callback function is present,
            or if the ``salinfo`` has not started reading.

        Notes
        -----
        Use with caution when mixing with `next`, since that also
        consumes data from the queue.
        """
        self.salinfo.assert_started()
        if self.has_callback:
            raise RuntimeError("Not allowed because there is a callback function")
        if self._data_queue:
            return self._data_queue.popleft()
        return None

    async def next(self, *, flush, timeout=None):
        """Pop and return the oldest message from the queue, waiting for data
        if the queue is empty.

        This method affects the data returned by `get_oldest`,
        but not the data returned by `aget` or `get`.

        Parameters
        ----------
        flush : `bool`
            If `True` then flush the queue before starting a read.
            This guarantees that the method will wait for a new message.
            If `False` and there is data on the queue, then pop and return
            the oldest message from the queue, without waiting;
            if queue is empty then wait for a new message.
        timeout : `float`, optional
            Time limit, in seconds. If None then no time limit.

        Returns
        -------
        data : `DataType`
            The message data.

        Raises
        ------
        RuntimeError
            If a callback function is present,
            or if the ``salinfo`` has not started reading.

        Notes
        -----
        Do not modify the returned data. To make a copy that you can
        safely modify, use ``copy.copy(data)``.
        """
        self.salinfo.assert_started()
        if self.has_callback:
            raise RuntimeError("Not allowed because there is a callback function")
        if flush:
            self.flush()
        return await self._next(timeout=timeout)

    async def _next(self, *, timeout=None):
        """Implement next.

        Unlike `next`, this can be called while using a callback function.
        """
        self.python_queue_length_checker.check_nitems(len(self._data_queue))
        if self._data_queue:
            return self._data_queue.popleft()
        if self._next_task.done():
            self._next_task = asyncio.Future()
        return await asyncio.wait_for(self._next_task, timeout=timeout)

    async def _callback_loop(self):
        while True:
            if not self.has_callback:
                return
            data = await self._next()
            result = self._run_callback(data)
            if self.allow_multiple_callbacks:
                # Purge done callback tasks and add a new one.
                self._callback_tasks = {
                    task for task in self._callback_tasks if not task.done()
                }
                self._callback_tasks.add(asyncio.create_task(result))
            else:
                await result

    def _cancel_callbacks(self):
        """Cancel the callback loop and all existing callback tasks."""
        self._callback_loop_task.cancel()
        while self._callback_tasks:
            task = self._callback_tasks.pop()
            task.cancel()

    async def _run_callback(self, data):
        try:
            result = self._callback(data)
            if inspect.isawaitable(result):
                await result
        except asyncio.CancelledError:
            raise
        except Exception as e:
            if not isinstance(e, base.ExpectedError):
                self.log.exception(f"Callback {self.callback} failed with data={data}")

    def _queue_data(self, data_list, loop):
        """Queue multiple one or more messages.

        Parameters
        ----------
        data_list : `list` [dds_messages]
            DDS messages to be queueued.
        loop : `asyncio.AbstractEventLoop` or `None`
            Foreground asyncio loop.
            Specify the loop if and only if running from a background thread.
            If running from the main asyncio loop specify ``loop = None``
            (this is done to read historical data while starting).

        Also update ``self._current_data`` and fire `self._next_task`
        (if pending).
        """
        if not data_list:
            return
        for data in data_list:
            self._queue_one_item(data)
        self._current_data = data
        if loop is not None and loop.is_running():
            # Reading messages in a background thread.
            loop.call_soon_threadsafe(self._report_next)
        else:
            # Reading messages in the main thread.
            self._report_next()

    def _queue_one_item(self, data):
        """Add a single message to the Python queue.

        Subclasses may override this to modify the message before queuing.
        `ControllerCommand` does this.
        """
        self._data_queue.append(data)

    def _report_next(self):
        """Set self._next_task to the oldest message on the queue.

        A no-op if self._next_task is done or the queue is empty.
        """
        if not self._next_task.done() and self._data_queue:
            oldest_message = self._data_queue.popleft()
            self._next_task.set_result(oldest_message)



try:
    from .version import *
except ImportError:
    __version__ = "?"

from .async_s3_bucket import *
from .sal_enums import *
from .base import *
from .dds_utils import *
from .idl_metadata import *
from .domain import *
from .sal_info import *
from .sal_log_handler import *
from .base_csc import *
from .validator import *
from .configurable_csc import *
from .controller import *
from .remote import *
from .base_script import *
from .csc_utils import *
from .testutils import *
from .config_schema import *
from .testcsc import *
from .testscript import *
from .base_csc_test_case import *
from .base_config_test_case import *
from .csc_commander import *
from .testcsccommander import *



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import contextlib
import itertools
import logging
import pathlib
import subprocess
import unittest
import warnings

import numpy as np

from lsst.ts import salobj

# Long enough to perform any reasonable operation
# including starting a CSC or loading a script (seconds)
STD_TIMEOUT = 60
# Timeout for when we expect no new data (seconds).
NODATA_TIMEOUT = 0.1

np.random.seed(47)

index_gen = salobj.index_generator()
TEST_DATA_DIR = TEST_CONFIG_DIR = pathlib.Path(__file__).resolve().parent / "data"
TEST_CONFIG_DIR = TEST_DATA_DIR / "config"


def all_permutations(items):
    """Return all permutations of a list of items and of all sublists,
    including [].
    """
    for i in range(len(items) + 1):
        for val in itertools.permutations(items, r=i):
            yield val


class FailInReportFaultCsc(salobj.TestCsc):
    """A Test CSC that fails in report_summary_state when reporting fault.

    This CSC always starts up in the ENABLED state.

    Used to test for infinite loops and other issues when something
    goes wrong while going to the fault state.

    Parameters
    ----------
    index : `int`
        SAL index.
    doraise : `bool`
        Raise in report_summary_state?
        If false then call fault instead.
    report_first : `bool`
        Call super().report_summary_state first in report_summary_state?
        If false then call it last.
    """

    def __init__(self, index, doraise, report_first):
        super().__init__(index=index, initial_state=salobj.State.ENABLED)
        self.doraise = doraise
        self.report_first = report_first

    def report_summary_state(self):
        if self.report_first:
            super().report_summary_state()
        if self.summary_state == salobj.State.FAULT:
            if self.doraise:
                raise RuntimeError(
                    "Intentionally raise an exception when going to the FAULT state"
                )
            else:
                self.fault(
                    code=10934,
                    report="a report that will be ignored",
                    traceback="a traceback that will be ignored",
                )
        if not self.report_first:
            super().report_summary_state()


class CommunicateTestCase(salobj.BaseCscTestCase, unittest.IsolatedAsyncioTestCase):
    def basic_make_csc(self, initial_state, config_dir, simulation_mode):
        return salobj.TestCsc(
            self.next_index(),
            initial_state=initial_state,
            config_dir=config_dir,
            simulation_mode=simulation_mode,
        )

    @contextlib.asynccontextmanager
    async def make_remote(
        self,
        identity,
    ):
        """Create a remote to talk to self.csc with a specified identity.

        Uses the domain created by make_csc.

        Parameters
        ----------
        identity : `str`
            Identity for remote.

        Notes
        -----
        Adds a logging.StreamHandler if one is not already present.
        """
        domain = self.csc.domain
        original_default_identity = domain.default_identity
        try:
            domain.default_identity = identity
            remote = salobj.Remote(
                domain=domain,
                name=self.csc.salinfo.name,
                index=self.csc.salinfo.index,
            )
        finally:
            domain.default_identity = original_default_identity
        self.assertEqual(remote.salinfo.identity, identity)
        try:
            await remote.start_task
            yield remote
        finally:
            await remote.close()

    async def test_authorization(self):
        """Test authorization.

        For simplicity this test calls setAuthList without a +/- prefix.
        The prefix is tested elsewhere.
        """
        async with self.make_csc(initial_state=salobj.State.ENABLED):
            # TODO DM-26605 remove this line as no longer needed:
            self.csc.authorize = True
            await self.assert_next_sample(
                self.remote.evt_authList,
                authorizedUsers="",
                nonAuthorizedCSCs="",
            )

            domain = self.csc.salinfo.domain

            # Note that self.csc and self.remote have the same user_host.
            csc_user_host = domain.user_host

            # Make a remote that pretends to be from a different CSC
            # and test non-authorized CSCs
            other_name_index = "Script:5"
            async with self.make_remote(identity=other_name_index) as other_csc_remote:

                all_csc_names = ["ATDome", "Hexapod:1", other_name_index]
                for csc_names in all_permutations(all_csc_names):
                    csc_names_str = ", ".join(csc_names)
                    with self.subTest(csc_names_str=csc_names_str):
                        await self.remote.cmd_setAuthList.set_start(
                            nonAuthorizedCSCs=csc_names_str, timeout=STD_TIMEOUT
                        )
                        await self.assert_next_sample(
                            self.remote.evt_authList,
                            authorizedUsers="",
                            nonAuthorizedCSCs=", ".join(sorted(csc_names)),
                        )
                        if other_name_index in csc_names:
                            # A blocked CSC; this should fail.
                            with salobj.assertRaisesAckError(
                                ack=salobj.SalRetCode.CMD_NOPERM
                            ):
                                await other_csc_remote.cmd_wait.set_start(
                                    duration=0, timeout=STD_TIMEOUT
                                )
                        else:
                            # Not a blocked CSC; this should work.
                            await other_csc_remote.cmd_wait.set_start(
                                duration=0, timeout=STD_TIMEOUT
                            )

                        # My user_host should work regardless of
                        # non-authorized CSCs.
                        await self.remote.cmd_wait.set_start(
                            duration=0, timeout=STD_TIMEOUT
                        )

                        # Disabling authorization should always work
                        self.csc.cmd_wait.authorize = False
                        try:
                            await other_csc_remote.cmd_wait.set_start(
                                duration=0, timeout=STD_TIMEOUT
                            )
                        finally:
                            self.csc.cmd_wait.authorize = True

            # Test authorized users that are not me.
            # Reported auth users should always be in alphabetical order;
            # test this by sending users NOT in alphabetical order.
            all_other_user_hosts = [f"notme{i}{csc_user_host}" for i in (3, 2, 1)]
            other_user_host = all_other_user_hosts[1]

            async with self.make_remote(identity=other_user_host) as other_user_remote:
                for auth_user_hosts in all_permutations(all_other_user_hosts):
                    users_str = ", ".join(auth_user_hosts)
                    with self.subTest(users_str=users_str):
                        await self.remote.cmd_setAuthList.set_start(
                            authorizedUsers=users_str,
                            nonAuthorizedCSCs="",
                            timeout=STD_TIMEOUT,
                        )
                        await self.assert_next_sample(
                            self.remote.evt_authList,
                            authorizedUsers=", ".join(sorted(auth_user_hosts)),
                            nonAuthorizedCSCs="",
                        )
                        if other_user_host in auth_user_hosts:
                            # An allowed user; this should work.
                            await other_user_remote.cmd_wait.set_start(
                                duration=0, timeout=STD_TIMEOUT
                            )
                        else:
                            # Not an allowed user; this should fail.
                            with salobj.assertRaisesAckError(
                                ack=salobj.SalRetCode.CMD_NOPERM
                            ):
                                await other_user_remote.cmd_wait.set_start(
                                    duration=0, timeout=STD_TIMEOUT
                                )

                        # Temporarily disable authorization and try again;
                        # this should always work.
                        self.csc.cmd_wait.authorize = False
                        try:
                            await other_user_remote.cmd_wait.set_start(
                                duration=0, timeout=STD_TIMEOUT
                            )
                        finally:
                            self.csc.cmd_wait.authorize = True

                        # My user_host should work regardless of
                        # authorized users.
                        self.remote.salinfo.domain.identity = csc_user_host
                        await self.remote.cmd_wait.set_start(
                            duration=0, timeout=STD_TIMEOUT
                        )

    async def test_set_auth_list_prefix(self):
        """Test the setAuthList command with a +/- prefix"""
        async with self.make_csc(initial_state=salobj.State.ENABLED):
            await self.assert_next_sample(
                self.remote.evt_authList,
                authorizedUsers="",
                nonAuthorizedCSCs="",
            )

            all_csc_names = ["Test:5", "ATDome:0", "ATDome"]
            expected_cscs_set = set()
            # Test adding non-authorized CSCs.
            # Test that a trailing :0 is stripped (so, in this test
            # setAuthList treats ATDome:0 and ATDome as the same CSC).
            # Reported non-auth CSCs should always be in alphabetical order;
            # test this by sending CSC names NOT in alphabetical order.
            for i in range(len(all_csc_names)):
                csc_names = all_csc_names[:i]
                # Compute a variant of csc_names with trailing :0 stripped
                # and use that for the expected output from the authList event.
                nonzero_csc_names = [
                    csc[:-2] if csc.endswith(":0") else csc for csc in csc_names
                ]
                expected_cscs_set |= set(nonzero_csc_names)
                add_some_str = "+ " + ", ".join(csc_names)
                await self.remote.cmd_setAuthList.set_start(
                    nonAuthorizedCSCs=add_some_str, timeout=STD_TIMEOUT
                )
                await self.assert_next_sample(
                    self.remote.evt_authList,
                    authorizedUsers="",
                    nonAuthorizedCSCs=", ".join(sorted(expected_cscs_set)),
                )

            # Removing CSCs should have the expected effect
            for i in range(len(all_csc_names)):
                csc_names = all_csc_names[:i]
                # Compute a variant of csc_names with trailing :0 stripped
                # and use that for the expected output from the authList event.
                nonzero_csc_names = [
                    csc[:-2] if csc.endswith(":0") else csc for csc in csc_names
                ]
                expected_cscs_set -= set(nonzero_csc_names)
                remove_some_str = "- " + ", ".join(csc_names)
                await self.remote.cmd_setAuthList.set_start(
                    nonAuthorizedCSCs=remove_some_str, timeout=STD_TIMEOUT
                )
                await self.assert_next_sample(
                    self.remote.evt_authList,
                    authorizedUsers="",
                    nonAuthorizedCSCs=", ".join(sorted(expected_cscs_set)),
                )

            # Test setting authorized users.
            # Test that user = csc_user_host (the CSC's user_host) is ignored.
            # Reported auth users should always be in alphabetical order;
            # test this by sending users NOT in alphabetical order.
            csc_user_host = salobj.get_user_host()
            all_users_hosts = [csc_user_host] + [
                f"notme{i}{csc_user_host}" for i in (3, 2, 1)
            ]
            expected_user_set = set()
            # Test adding authorized users; only users other than "me"
            # are actually added.
            for i in range(len(all_users_hosts)):
                user_names = all_users_hosts[:i]
                expected_user_set |= set(user_names)
                expected_user_set -= {csc_user_host}
                add_some_str = "+ " + ", ".join(user_names)
                await self.remote.cmd_setAuthList.set_start(
                    authorizedUsers=add_some_str, timeout=STD_TIMEOUT
                )
                await self.assert_next_sample(
                    self.remote.evt_authList,
                    authorizedUsers=", ".join(sorted(expected_user_set)),
                    nonAuthorizedCSCs="",
                )

            # Removing authorized users should have the expected effect
            for i in range(len(all_users_hosts)):
                user_names = all_users_hosts[:i]
                expected_user_set -= set(user_names)
                remove_some_str = "- " + ", ".join(user_names)
                await self.remote.cmd_setAuthList.set_start(
                    authorizedUsers=remove_some_str, timeout=STD_TIMEOUT
                )
                await self.assert_next_sample(
                    self.remote.evt_authList,
                    authorizedUsers=", ".join(sorted(expected_user_set)),
                    nonAuthorizedCSCs="",
                )

    async def test_heartbeat(self):
        async with self.make_csc(initial_state=salobj.State.STANDBY):
            self.csc.heartbeat_interval = 0.1
            await self.remote.evt_heartbeat.next(flush=True, timeout=STD_TIMEOUT)
            await self.remote.evt_heartbeat.next(flush=True, timeout=0.2)
            await self.remote.evt_heartbeat.next(flush=True, timeout=0.2)
            await self.remote.evt_heartbeat.next(flush=True, timeout=0.2)

    async def test_bin_script_run(self):
        """Test running the Test CSC from the bin script.

        Note that the bin script calls class method ``amain``.
        """
        for initial_state, settings_to_apply in (
            (None, None),
            (salobj.State.STANDBY, None),
            (salobj.State.DISABLED, "all_fields.yaml"),
            (salobj.State.ENABLED, ""),
        ):
            index = self.next_index()
            with self.subTest(
                initial_state=initial_state, settings_to_apply=settings_to_apply
            ):
                await self.check_bin_script(
                    name="Test",
                    index=index,
                    exe_name="run_test_csc.py",
                    initial_state=initial_state,
                    settings_to_apply=settings_to_apply,
                )

    async def test_bin_script_version(self):
        """Test running the Test CSC from the bin script.

        Note that the bin script calls class method ``amain``.
        """
        salobj.set_random_lsst_dds_partition_prefix()
        index = self.next_index()
        exec_path = pathlib.Path(__file__).parents[1] / "bin" / "run_test_csc.py"

        process = await asyncio.create_subprocess_exec(
            str(exec_path),
            str(index),
            "--version",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        try:
            stdout, stderr = await asyncio.wait_for(
                process.communicate(), timeout=STD_TIMEOUT
            )
            self.assertEqual(stdout.decode()[:-1], salobj.__version__)
            await asyncio.wait_for(process.wait(), timeout=STD_TIMEOUT)
            self.assertEqual(process.returncode, 0)
        finally:
            if process.returncode is None:
                process.terminate()
                warnings.warn("Killed a process that was not properly terminated")

    async def test_log_level(self):
        """Test that specifying a log level to make_csc works."""
        # If specified then log level is the value given.
        async with self.make_csc(
            initial_state=salobj.State.STANDBY, log_level=logging.DEBUG
        ):
            self.assertEqual(self.csc.log.getEffectiveLevel(), logging.DEBUG)

        async with self.make_csc(
            initial_state=salobj.State.STANDBY, log_level=logging.WARNING
        ):
            self.assertEqual(self.csc.log.getEffectiveLevel(), logging.WARNING)

        # At this point log level is WARNING; now check that by default
        # log verbosity is increased (log level decreased) to INFO.
        async with self.make_csc(initial_state=salobj.State.STANDBY):
            self.assertEqual(self.csc.log.getEffectiveLevel(), logging.INFO)

    async def test_setArrays_command(self):
        async with self.make_csc(initial_state=salobj.State.ENABLED):
            # until the controller gets its first setArrays
            # it will not send any arrays events or telemetry
            self.assertFalse(self.csc.evt_arrays.has_data)
            self.assertFalse(self.csc.tel_arrays.has_data)
            self.assertFalse(self.remote.evt_arrays.has_data)
            self.assertFalse(self.remote.tel_arrays.has_data)
            self.assertIsNone(self.remote.evt_arrays.get())
            self.assertIsNone(self.remote.tel_arrays.get())

            # send the setArrays command with random data
            cmd_data_sent = self.csc.make_random_cmd_arrays()
            await self.remote.cmd_setArrays.start(cmd_data_sent, timeout=STD_TIMEOUT)

            # see if new data was broadcast correctly
            evt_data = await self.remote.evt_arrays.next(
                flush=False, timeout=STD_TIMEOUT
            )
            self.csc.assert_arrays_equal(cmd_data_sent, evt_data)
            tel_data = await self.remote.tel_arrays.next(
                flush=False, timeout=STD_TIMEOUT
            )
            self.csc.assert_arrays_equal(cmd_data_sent, tel_data)

            self.assertTrue(self.csc.evt_arrays.has_data)
            self.assertTrue(self.csc.tel_arrays.has_data)
            self.assertTrue(self.remote.evt_arrays.has_data)
            self.assertTrue(self.remote.tel_arrays.has_data)

            # also test get
            self.csc.assert_arrays_equal(cmd_data_sent, self.remote.tel_arrays.get())
            self.csc.assert_arrays_equal(cmd_data_sent, self.remote.evt_arrays.get())

    async def test_setScalars_command(self):
        async with self.make_csc(initial_state=salobj.State.ENABLED):
            # until the controller gets its first setArrays
            # it will not send any arrays events or telemetry
            self.assertFalse(self.csc.evt_scalars.has_data)
            self.assertFalse(self.csc.tel_scalars.has_data)
            self.assertFalse(self.remote.evt_scalars.has_data)
            self.assertFalse(self.remote.tel_scalars.has_data)
            self.assertIsNone(self.remote.evt_scalars.get())
            self.assertIsNone(self.remote.tel_scalars.get())

            # send the setScalars command with random data
            cmd_data_sent = self.csc.make_random_cmd_scalars()
            await self.remote.cmd_setScalars.start(cmd_data_sent, timeout=STD_TIMEOUT)

            # see if new data is being broadcast correctly
            evt_data = await self.remote.evt_scalars.next(
                flush=False, timeout=STD_TIMEOUT
            )
            self.csc.assert_scalars_equal(cmd_data_sent, evt_data)
            tel_data = await self.remote.tel_scalars.next(
                flush=False, timeout=STD_TIMEOUT
            )
            self.csc.assert_scalars_equal(cmd_data_sent, tel_data)

            self.assertTrue(self.csc.evt_scalars.has_data)
            self.assertTrue(self.csc.tel_scalars.has_data)
            self.assertTrue(self.remote.evt_scalars.has_data)
            self.assertTrue(self.remote.tel_scalars.has_data)

            # also test get
            self.csc.assert_scalars_equal(cmd_data_sent, self.remote.tel_scalars.get())
            self.csc.assert_scalars_equal(cmd_data_sent, self.remote.evt_scalars.get())

    async def test_fault_state_transitions(self):
        """Test CSC state transitions into fault and out again.

        Going into the fault state is done via the ``fault`` command.
        """
        for initial_state in salobj.State:
            if initial_state == salobj.State.OFFLINE:
                # Not a valid initial state
                continue
            if initial_state == salobj.State.FAULT:
                # The ``fault`` command does nothing if TestCsc
                # is already in the FAULT state
                continue
            with self.subTest(initial_state=initial_state):
                async with self.make_csc(initial_state=initial_state):
                    await self.assert_next_summary_state(initial_state)

                    # Issue the ``fault`` command
                    # and check the state and error code.
                    await self.remote.cmd_fault.start(timeout=STD_TIMEOUT)
                    await self.assert_next_summary_state(salobj.State.FAULT)
                    await self.assert_next_sample(
                        topic=self.remote.evt_errorCode, errorCode=1
                    )

                    # Issue the ``standby`` command to recover.
                    await self.remote.cmd_standby.start(timeout=STD_TIMEOUT)
                    await self.assert_next_summary_state(salobj.State.STANDBY)

    async def test_fault_method(self):
        """Test BaseCsc.fault with and without optional arguments."""
        async with self.make_csc(initial_state=salobj.State.STANDBY):
            await self.assert_next_summary_state(salobj.State.STANDBY)
            with self.assertRaises(asyncio.TimeoutError):
                await self.remote.evt_errorCode.next(
                    flush=False, timeout=NODATA_TIMEOUT
                )

            code = 52
            report = "Report for error code"
            traceback = "Traceback for error code"

            # if an invalid code is specified then errorCode is not output
            # but the CSC stil goes into a FAULT state
            self.csc.fault(code="not a valid code", report=report)
            await self.assert_next_summary_state(salobj.State.FAULT)
            with self.assertRaises(asyncio.TimeoutError):
                await self.remote.evt_errorCode.next(
                    flush=False, timeout=NODATA_TIMEOUT
                )

            await self.remote.cmd_standby.start(timeout=STD_TIMEOUT)
            await self.assert_next_summary_state(salobj.State.STANDBY)

            # if code is specified then errorReport is output;
            # first test with report and traceback specified,
            # then without, to make sure those values are not cached
            self.csc.fault(code=code, report=report, traceback=traceback)
            await self.assert_next_summary_state(salobj.State.FAULT)
            await self.assert_next_sample(
                topic=self.remote.evt_errorCode,
                errorCode=code,
                errorReport=report,
                traceback=traceback,
            )

            # Try a disallowed command and check that the error report
            # is part of the traceback.
            with salobj.assertRaisesAckError(result_contains=report):
                await self.remote.cmd_wait.set_start(duration=5, timeout=STD_TIMEOUT)

            await self.remote.cmd_standby.start(timeout=STD_TIMEOUT)
            await self.assert_next_summary_state(salobj.State.STANDBY)

            self.csc.fault(code=code, report="")
            await self.assert_next_summary_state(salobj.State.FAULT)
            await self.assert_next_sample(
                topic=self.remote.evt_errorCode,
                errorCode=code,
                errorReport="",
                traceback="",
            )

            await self.remote.cmd_standby.start(timeout=STD_TIMEOUT)
            await self.remote.cmd_exitControl.start(timeout=STD_TIMEOUT)

    async def test_fault_problems(self):
        """Test BaseCsc.fault when report_summary_state misbehaves."""
        index = self.next_index()
        for doraise, report_first in itertools.product((False, True), (False, True)):
            with self.subTest(doraise=doraise, report_first=report_first):
                async with FailInReportFaultCsc(
                    index=index, doraise=doraise, report_first=report_first
                ) as csc, salobj.Remote(
                    domain=csc.domain, name="Test", index=index
                ) as remote:
                    await self.assert_next_summary_state(
                        salobj.State.ENABLED, remote=remote
                    )

                    code = 51
                    report = "Report for error code"
                    traceback = "Traceback for error code"
                    csc.fault(code=code, report=report, traceback=traceback)

                    await self.assert_next_summary_state(
                        salobj.State.FAULT, remote=remote
                    )
                    await self.assert_next_sample(
                        topic=remote.evt_errorCode,
                        errorCode=code,
                        errorReport=report,
                        traceback=traceback,
                    )

                    # make sure FAULT state and errorCode are only sent once
                    with self.assertRaises(asyncio.TimeoutError):
                        await remote.evt_summaryState.next(
                            flush=False, timeout=NODATA_TIMEOUT
                        )
                    with self.assertRaises(asyncio.TimeoutError):
                        await remote.evt_errorCode.next(
                            flush=False, timeout=NODATA_TIMEOUT
                        )

    async def test_make_csc_timeout(self):
        """Test that setting the timeout argument to make_csc works."""
        with self.assertRaises(asyncio.TimeoutError):
            # Use such a short timeout that make_csc times out
            async with self.make_csc(initial_state=salobj.State.STANDBY, timeout=0):
                pass

    async def test_standard_state_transitions(self):
        """Test standard CSC state transitions.

        The initial state is STANDBY.
        The standard commands and associated state transitions are:

        * start: STANDBY to DISABLED
        * enable: DISABLED to ENABLED

        * disable: ENABLED to DISABLED
        * standby: DISABLED or FAULT to STANDBY
        * exitControl: STANDBY to OFFLINE (quit)
        """
        async with self.make_csc(
            initial_state=salobj.State.STANDBY, config_dir=TEST_CONFIG_DIR
        ):
            await self.check_standard_state_transitions(
                enabled_commands=("setArrays", "setScalars", "wait"),
                skip_commands=("fault",),
                settingsToApply="all_fields",
            )


if __name__ == "__main__":
    unittest.main()



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["Domain"]

import asyncio
import os
import weakref
import warnings

import dds

from lsst.ts import ddsconfig
from lsst.ts import idl
from . import base

MAX_RANDOM_HOST = (1 << 31) - 1


class QosSet:
    """QoS profiles for topic, reader and writer.

    Parameters
    ----------
    qos_path : `str` or `pathlib.Path`
        Path to QoS XML file.
    profile_name : `str`
        Profile name; one of "Command", "Event", or "Telemetry".

    Attributes
    ----------
    profile_name : `str`
        Profile name; one of "Command", "Event", or "Telemetry".
    qos_provider : `dds.QosProvider`
        QoS provider.
    topic_qos : `dds.Qos`
        Topic QoS
    reader_qos : `dds.Qos`
        Topic reader QoS
    writer_qos : `dds.Qos`
        Topic writer QoS

    Notes
    -----
    The following QoS should be created elsewhere:
    * publisher and subscriber: these depend on the DDS partition name,
      which is only read when creating the SalInfo object.
      Also these QoS are the same for all QoS profiles.
    * participant: this is the same for all QoS profiles,
      so there is no point making 3 of them.
    """

    def __init__(self, qos_path, profile_name):
        self.profile_name = profile_name
        self.qos_provider = dds.QosProvider(qos_path.as_uri(), profile_name)
        self.topic_qos = self.qos_provider.get_topic_qos()
        self.reader_qos = self.qos_provider.get_reader_qos()
        self.writer_qos = self.qos_provider.get_writer_qos()

    @property
    def volatile(self):
        """Does this category of topics have volatile durability?

        Volatile topics provide no late-joiner data.
        """
        return self.topic_qos.durability.kind == dds.DDSDurabilityKind.VOLATILE


class Domain:
    r"""dds domain participant and quality of service information.

    Attributes
    ----------
    participant : ``dds.DomainParticipant``
        DDS domain participant.
    origin : `int`
        Process ID. Used to set the ``private_origin`` field of output samples.
    default_identity : `str`
        Default value used for the identity field of `SalInfo`.
        Initalized to ``user_host`` but `Controller`\ 's constructor
        sets it to `SalInfo.user_index` so that all `Remote`\ s
        constructed with the controller's domain will have the
        controller's identity.
        For testing purposes, it is allowed to change this field
        before constructing a `Remote`.
    user_host : `str`
        username@host. This will match ``identity`` unless the latter
        is set to a CSC name.
    idl_dir : `pathlib.Path`
        Root directory of the ``ts_idl`` package.
    ackcmd_qos_set : `QosSet`
        QoS set for the ackcmd topic.
    command_qos_set : `QosSet`
        QoS set for command topics.
    event_qos_set : `QosSet`
        QoS set for event topics.
    telemetry_qos_set : `QosSet`
        QoS set for telemetry topics.

    Notes
    -----
    Reads the following `Environment Variables
    <https://ts-salobj.lsst.io/configuration.html#environment_variables>`_;
    follow the link for details:

    * OSPL_MASTER_PRIORITY, optional is used to set the Master Priority.
      If present, it must be a value between 0 and 255.

    **Cleanup**

    It is important to close a `Domain` when you are done with it, especially
    in unit tests, because otherwise unreleased resources may cause problems.
    To make this easier `Domain` is a context manager.

    `Domain` contains a registry of `SalInfo` instances, so that it can close
    them when the domain is closed. The registry uses weak references to avoid
    circular dependencies (since SalInfo stores the domain). `SalInfo`
    instances automatically register themselves with the provided domain.

    To conserve resources you should have only one `Domain` per process.
    Thus:

    * `Controller` creates its own Domain as attribute ``domain``
      and closes it when the controller is closed. To create a
      `Remote` for a controller, use the domain from the controller, e.g.::

        controller = Controller(name="Test", index=47)
        try:
            remote = Remote(domain=controller.domain, name="Test", index=47)
            ...
        finally:
            await controller.close()

        # or if you only want the objects for the duration of a function,
        # e.g in a unit test:
        async with Controller(name="Test", index=47) as controller:
            remote = Remote(domain=controller.domain, name="Test", index=47)
            ...

    * If you are creating `Remote`\ s without a `Controller` then you must
      create the Domain yourself and close it when you are finished, e.g.::

        domain = salobj.Domain()
        try:
            remote = salobj.Remote(domain=domain, name="Test", index=47)
            ...
        finally:
            await domain.close()

        # or if you only want the objects for the duration of a function,
        # e.g in a unit test:
        async with salobj.Domain() as domain:
            test_remote = salobj.Remote(domain=domain, name="Test", index=5)
    """

    def __init__(self):
        self.isopen = True
        self.user_host = base.get_user_host()
        self.default_identity = self.user_host

        # Accumulators for verifying that close is working.
        self.num_read_loops = 0

        self.done_task = asyncio.Future()

        # Set of SalInfo.
        self._salinfo_set = weakref.WeakSet()

        self.origin = os.getpid()
        self.idl_dir = idl.get_idl_dir()

        qos_path = ddsconfig.get_qos_path()
        self.ackcmd_qos_set = QosSet(qos_path=qos_path, profile_name="AckcmdProfile")
        self.command_qos_set = QosSet(qos_path=qos_path, profile_name="CommandProfile")
        self.event_qos_set = QosSet(qos_path=qos_path, profile_name="EventProfile")
        self.telemetry_qos_set = QosSet(
            qos_path=qos_path, profile_name="TelemetryProfile"
        )

        # Any of the three qos providers is fine for the participant qos.
        participant_qos = self.command_qos_set.qos_provider.get_participant_qos()
        self.participant = dds.DomainParticipant(qos=participant_qos)

    @property
    def salinfo_set(self):
        return self._salinfo_set

    def add_salinfo(self, salinfo):
        """Add the specified salinfo to the internal registry.

        Parameters
        ----------
        salinfo : `SalInfo`
            SAL component information

        Raises
        ------
        RuntimeError
            If the salinfo is already present
        """
        if salinfo in self._salinfo_set:
            raise RuntimeError(f"salinfo {salinfo} already added")
        self._salinfo_set.add(salinfo)

    def make_publisher(self, partition_names):
        """Make a dds publisher.

        Parameters
        ----------
        partition_names : `list` [`str`]
            List of DDS partition names.
        """
        partition_qos_policy = dds.PartitionQosPolicy(partition_names)

        # Any qos set will do, because the publisher and subscriber QoS
        # is the same for all of them.
        publisher_qos = self.event_qos_set.qos_provider.get_publisher_qos()
        publisher_qos.set_policies([partition_qos_policy])

        return self.participant.create_publisher(publisher_qos)

    def make_subscriber(self, partition_names):
        """Make a dds subscriber.

        Parameters
        ----------
        partition_names : `list` [`str`]
            List of DDS partition names.
        """
        partition_qos_policy = dds.PartitionQosPolicy(partition_names)

        # Any qos set will do, because the publisher and subscriber QoS
        # is the same for all of them.
        subscriber_qos = self.event_qos_set.qos_provider.get_subscriber_qos()
        subscriber_qos.set_policies([partition_qos_policy])

        return self.participant.create_subscriber(subscriber_qos)

    def remove_salinfo(self, salinfo):
        """Remove the specified salinfo from the internal registry.

        Parameters
        ----------
        salinfo : `SalInfo`
            SAL component information

        Returns
        -------
        removed : `boolean`
            True if removed, False if not found
        """
        try:
            self._salinfo_set.remove(salinfo)
            return True
        except KeyError:
            return False

    def basic_close(self):
        """A synchronous and less thorough version of `close`.

        Intended for exit handlers and constructor error handlers.
        """
        try:
            while self._salinfo_set:
                salinfo = self._salinfo_set.pop()
                salinfo.basic_close()
        finally:
            self.participant.close()

    async def close(self):
        """Close all registered `SalInfo` and the dds domain participant.

        May be called multiple times. The first call closes the Domain;
        subsequent calls wait until the Domain is closed.
        """
        if not self.isopen:
            await self.done_task
            return
        self.isopen = False
        try:
            while self._salinfo_set:
                salinfo = self._salinfo_set.pop()
                await salinfo.close()
        finally:
            self.participant.close()
        if self.num_read_loops != 0:
            warnings.warn(
                f"After Domain.close num_read_loops={self.num_read_loops}; it should be 0"
            )
        self.done_task.set_result(None)

    async def __aenter__(self):
        return self

    async def __aexit__(self, type, value, traceback):
        await self.close()



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["DefaultingValidator"]

import jsonschema


class DefaultingValidator:
    """A wrapper for jsonschema validators that applies default values.

    Parameters
    ----------
    schema : `dict`
        Schema against which to validate.
    ValidatorClass : ``jsonschema.IValidator``, optional
        jsonschema validator class, e.g. ``jsonschema.Draft7Validator``.

    Notes
    -----
    Default values are handled at most 2 levels deep in an object hierarchy.
    For deeper hierarchies, set the default at a higher level.
    For example::

        type: object
        properties:
          number1:
            type: number
            default: 1
          subdict1:
            type: object
            properties:
              number2:
                type: number
                default: 2
              subdict2:
                type: object
                properties:
                  number3:
                    type: number
                    # default is ignored this deep; set it at a higher level
                default:
                  number3: 3

    This class is not a ``jsonschema.IValidator`` but it contains two
    validators:

    * defaults_validator: a validator that sets default values in the
      data being validated
    * final_validator: a standard validator that does not alter
      the data being validated.
    """

    def __init__(self, schema, ValidatorClass=jsonschema.Draft7Validator):
        ValidatorClass.check_schema(schema)
        self.final_validator = ValidatorClass(schema=schema)

        validate_properties = ValidatorClass.VALIDATORS["properties"]

        def set_defaults(validator, properties, instance, schema):
            """Wrap a jsonschema Validator so that it sets default values.

            Parameters
            ----------
            validator : ``jsonschema.IValidator``
                jsonschema validator.
            properties : `dict`
                The value of the property being validated within the instance
            instance : `dict`
                The item being checked and possibly set.
            schema : `dict`
                The schema being validated.

            Notes
            -----
            This code is based on https://python-jsonschema.readthedocs.io/
                en/stable/faq/#why-doesn-t-my-schema-
                s-default-property-set-the-default-on-my-instance
            but I added skip_properties to avoid infinite recursion
            """
            # most of these items cause infinite recursion if allowed through
            # and none are needed for setting defaults
            skip_properties = set(
                (
                    "additionalItems",
                    "additionalProperties",
                    "definitions",
                    "default",
                    "items",
                    "patternProperties",
                    "property",
                    "properties",
                    "readOnly",
                    "uniqueItems",
                )
            )
            for prop, subschema in properties.items():
                if not isinstance(subschema, dict):
                    continue
                if not isinstance(instance, dict):
                    continue
                if prop in skip_properties:
                    continue
                if "default" in subschema:
                    instance.setdefault(prop, subschema["default"])
                elif subschema.get("type") == "object" and "properties" in subschema:
                    # Handle defaults for one level deep sub-object.
                    subdefaults = {}
                    for subpropname, subpropvalue in subschema["properties"].items():
                        if "default" in subpropvalue:
                            subdefaults[subpropname] = subpropvalue["default"]
                    if subdefaults:
                        instance.setdefault(prop, subdefaults)

            for error in validate_properties(validator, properties, instance, schema):
                yield error

        WrappedValidator = jsonschema.validators.extend(
            ValidatorClass, {"properties": set_defaults}
        )
        WrappedValidator.check_schema(schema)
        self.defaults_validator = WrappedValidator(schema=schema)

    def validate(self, data_dict):
        """Validate data.

        Set missing values based on defaults in the schema,
        then check the final result against the schema
        (in case any defaults are not valid).

        Parameters
        ----------
        data_dict : `dict` or `None`
            Data to validate. If None then an empty dict is used.

        Returns
        -------
        result : `dict`
            Validated data. A copy of data_dict with missing values
            that have defaults set to those defaults.

        Raises
        ------
        jsonschema.exceptions.ValidationError
            If the data does not match the schema.
        """
        if data_dict is None:
            result = {}
        elif type(data_dict) is not dict:
            raise jsonschema.exceptions.ValidationError(f"{data_dict} is not a dict")
        else:
            result = data_dict.copy()

        self.defaults_validator.validate(result)
        self.final_validator.validate(result)
        return result
